from jaseci.utils.test_core import CoreTest


class UserApiTest(CoreTest):
    """Unit tests for Jac Walker APIs"""

    fixture_src = __file__

    def test_user_create_global_init(self):
        self.call(
            self.smast,
            ["sentinel_register", {"code": self.load_jac("hello_world.jac")}],
        )
        self.call(
            self.smast,
            ["global_sentinel_set", {}],
        )
        ret = self.call(
            self.mast,
            [
                "user_create",
                {
                    "name": "joe",
                    "global_init": "init",
                    "global_init_ctx": {"wrld": "jason"},
                },
            ],
        )
        self.assertEqual(
            ret["global_init"]["auto_run_result"]["report"], ["hello jason!"]
        )
        self.assertEqual(ret["user"]["name"], "joe")



"""
Admin Global api functions as a mixin
"""
from jaseci.api.interface import Interface
import jaseci.actions.live_actions as lact
import json


class ActionsApi:
    """
    APIs to manage actions

    This set action APIs enable the manual management of Jaseci actions and action
    libraries/sets. Action libraries can be loaded locally into the running instance of
    the python program, or as a remote container linked action library. In this mode,
    action libraries operate as micro-services. Jaseci will be able to dynamically
    and automatically make this decision for the user based on online monitoring and
    performance profiling.
    """

    @Interface.admin_api(cli_args=["file"])
    def actions_load_local(self, file: str):
        """
        Hot load a python module and assimilate any Jaseci Actions

        This API will dynamically load a module based on a python file. The module
        is loaded directly into the running Jaseci python instance. This API also
        makes an attempt to auto detect and hot load any python package dependencies
        the file may reference via python's relative imports. This file is assumed to
        have the necessary annotations and decorations required by Jaseci to recognize
        its actions.

        :param file: The python file with full to load actions from.
            (i.e., ~/local/myact.py)
        """
        success = lact.load_local_actions(file)
        if success:
            cur_config = self.config_get("ACTION_SETS")
            if cur_config and (not isinstance(cur_config, list)):
                config = json.loads(cur_config)
                if file not in config["local"]:
                    config["local"].append(file)
                    self.config_set("ACTION_SETS", json.dumps(config))
            else:
                self.config_set(
                    "ACTION_SETS",
                    json.dumps({"local": [file], "remote": [], "module": []}),
                )
        return {"success": success}

    @Interface.admin_api(cli_args=["url"])
    def actions_load_remote(self, url: str):
        """
        Hot link to a container linked action library

        This API will dynamically load a set of actions that are present on a remote
        server/micro-service. This server must be configured to interact with Jaseci
        properly. This is easily achieved using the same decorators used for local
        action libraries. Remote actions allow for higher flexibility in the languages
        supported for action libraries. If an  library writer would like to use another
        language, the main hook REST api simply needs to be implemented. Please
        refer to documentation on creating action libraries for more details.

        :param url: The url of the API server supporting Jaseci actions.
        """
        success = lact.load_remote_actions(url)
        if success:
            cur_config = self.config_get("ACTION_SETS")
            if cur_config and (not isinstance(cur_config, list)):
                config = json.loads(cur_config)
                if url not in config["remote"]:
                    config["remote"].append(url)
                    self.config_set("ACTION_SETS", json.dumps(config))
            else:
                self.config_set(
                    "ACTION_SETS",
                    json.dumps({"local": [], "remote": [url], "module": []}),
                )
        return {"success": success}

    @Interface.admin_api(cli_args=["mod"])
    def actions_load_module(self, mod: str):
        """
        Hot load a python module and assimilate any Jaseci Actions

        This API will dynamically load a module using python's module import format.
        This is particularly useful for pip installed action libraries as the developer
        can directly reference the module using the same format as a regular python
        import. As with load local, the module will be loaded directly into the running
        Jaseci python instance.

        :param mod: The import style module to load actions from.
            (i.e., jaseci_ai_kit.bi_enc)
        """
        success = lact.load_module_actions(mod)
        if success:
            cur_config = self.config_get("ACTION_SETS")
            if cur_config and (not isinstance(cur_config, list)):
                config = json.loads(cur_config)
                if mod not in config["module"]:
                    config["module"].append(mod)
                    self.config_set("ACTION_SETS", json.dumps(config))
            else:
                self.config_set(
                    "ACTION_SETS",
                    json.dumps({"local": [], "remote": [], "module": [mod]}),
                )
        return {"success": success}

    # @interface.admin_api()
    # def actions_get(self, name: str, value: str):
    #     """
    #     """

    @Interface.admin_api()
    def actions_list(self, name: str = ""):
        """
        List a set of or all loaded jaseci actions

        This API is used to list the loaded actions active in Jaseci. These actions
        include all types of loaded actions whether it be local modules or remote
        containers. A particular set of actions can be viewed using the name parameter.

        :param name: The name for a library for which to filter the view of shown
            actions. If left blank all actions from all loaded sets will be shown.
        """
        actions = list(lact.live_actions.keys())
        if len(name):
            actions = list(filter(lambda a: a.startswith(name), actions))
        return actions

    # @interface.admin_api()
    # def actions_delete(self, name: str, value: str):
    #     """
    #     """



"""
Queue api functions as a mixin
"""
from jaseci.api.interface import Interface


class QueueApi:
    """
    Queue APIs

    APIs used for celery configuration and monitoring
    """

    @Interface.private_api(allowed_methods=["get"])
    def walker_queue_check(self, task_id: str = ""):
        """
        Monitor Queues
        """
        if not self._h.task.is_running():
            return "Task hook is not yet initialized!"

        if not task_id:
            return self._h.task.inspect_tasks()
        else:
            return self._h.task.get_by_task_id(task_id)

    @Interface.private_api(allowed_methods=["get"])
    def walker_queue_wait(self, task_id: str):
        """
        Wait Queues
        """
        if not self._h.task.is_running():
            return "Task hook is not yet initialized!"

        if not task_id:
            return "Task id is required!"
        else:
            return self._h.task.get_by_task_id(task_id, True)



"""
Node class for Jaseci

Each node has an id, name, timestamp and it's set of edges.
First node in list of 'member_node_ids' is designated root node
"""
from jaseci.element.element import Element
from jaseci.element.obj_mixins import Anchored
from jaseci.graph.edge import Edge
from jaseci.utils.id_list import IdList
from jaseci.utils.utils import logger

import uuid


class Node(Element, Anchored):
    """Node class for Jaseci"""

    def __init__(self, dimension=0, **kwargs):
        self.edge_ids = IdList(self)
        self.parent_node_ids = IdList(self)
        self.member_node_ids = IdList(self)
        self.dimension = dimension  # Nodes are always hdgd 0
        Element.__init__(self, **kwargs)
        Anchored.__init__(self)

    def attach(self, node_obj, edge_set=None, as_outbound=True, as_bidirected=False):
        """
        Generalized attach function for attaching nodes with edges
        """
        if edge_set is None:
            edge_set = [
                Edge(
                    m_id=self._m_id,
                    h=self._h,
                    kind="edge",
                    name="generic",
                )
            ]
        link_order = [self, node_obj] if as_outbound else [node_obj, self]
        for e in edge_set:
            if not e.set_from_node(link_order[0]) or not e.set_to_node(link_order[1]):
                # Node not found error logged in set node function
                return []
            e.set_bidirected(as_bidirected)
        # save and return
        self.save()
        node_obj.save()
        return edge_set

    def attach_outbound(self, node_obj, edge_set=None):
        """
        Creates edge to a node and returns the edge
        edge_set is the list of edges to be used to make connections
        new edge is created if edge_set is empty
        """
        return self.attach(node_obj, edge_set, as_outbound=True)

    def attach_inbound(self, node_obj, edge_set=None):
        """
        Creates edge from a node and returns the edge
        edge_set is the list of edges to be used to make connections
        new edge is created if edge_set is empty
        """
        return self.attach(node_obj, edge_set, as_outbound=False)

    def attach_bidirected(self, node_obj, edge_set=None):
        """
        Creates bidirected edge to node returns the edge
        edge_set is the list of edges to be used to make connections
        new edge is created if edge_set is empty
        """
        return self.attach(node_obj, edge_set, as_bidirected=True)

    def detach(
        self,
        node_obj,
        edge_set=None,
        as_outbound=True,
        as_bidirected=False,
        ignore_direction=False,
        silent=True,
    ):
        """
        Generalized detach function for detaching nodes with edges
        """
        if edge_set is None:
            edge_set = self.attached_edges(node_obj)
        link_order = [self, node_obj] if as_outbound else [node_obj, self]
        num_detached = 0
        for e in edge_set:
            # validate edge connection exists
            if not e.connects(
                link_order[0], link_order[1], ignore_direction=ignore_direction
            ):
                if not silent:
                    logger.warning(
                        str(
                            f"{e} does not connect "
                            f"{link_order[0]} to {link_order[1]}"
                        )
                    )
                continue
            if as_bidirected and not ignore_direction and not e.is_bidirected():
                if not silent:
                    logger.warning(str(f"{e} is not a bidirected edge "))
                continue
            # destroy edge
            num_detached += 1
            e.destroy()
        return num_detached

    def detach_outbound(self, node_obj, edge_set=None, silent=True):
        """
        Destroy edges to a node
        edge_set is the list of edges to be detached (and distroyed)
        all edges are deteached and destroyed if edge_set empty
        returns number of detachments
        """
        return self.detach(node_obj, edge_set, as_outbound=True)

    def detach_inbound(self, node_obj, edge_set=None, silent=True):
        """
        Destroy edges from a node
        edge_set is the list of edges to be detached (and distroyed)
        all edges are deteached and destroyed if edge_set empty
        returns number of detachments
        """
        return self.detach(node_obj, edge_set, as_outbound=False)

    def detach_bidirected(self, node_obj, edge_set=None, silent=True):
        """
        Destroy bidirected edges between nodes
        edge_set is the list of edges to be detached (and distroyed)
        all edges are deteached and destroyed if edge_set empty
        returns number of detachments
        """
        return self.detach(node_obj, edge_set, as_bidirected=True)

    def detach_edges(self, node_obj, edge_set=None, silent=True):
        """
        Destroy given edges between nodes without checking orientation
        edge_set is the list of edges to be detached (and distroyed)
        all edges are deteached and destroyed if edge_set empty
        returns number of detachments
        """
        return self.detach(node_obj, edge_set, ignore_direction=True)

    def destroy_outbound(self, node_obj, edge_set=None):
        """
        Destroys attached node and all relevant edges
        Node and all edges are destroyed if edge_set empty
        """
        if edge_set is None:
            edge_set = self.outbound_edges(node_obj)
        if self.detach_outbound(node_obj, edge_set):
            node_obj.destroy()

    def destroy_inbound(self, node_obj, edge_set=None):
        """
        Destroys attached node and all relevant edges
        Node and all edges are destroyed if edge_set empty
        """
        if edge_set is None:
            edge_set = self.inbound_edges(node_obj)
        if self.detach_inbound(node_obj, edge_set):
            node_obj.destroy()

    def destroy_bidirected(self, node_obj, edge_set=None):
        """
        Destroys attached node and all relevant edges
        Node and all edges are destroyed if edge_set empty
        """
        if edge_set is None:
            edge_set = self.bidirected_edges(node_obj)
        if self.detach_bidirected(node_obj, edge_set):
            node_obj.destroy()

    def is_attached_out(self, node_obj, edge_set=None):
        """
        Tests whether edges attach to a node
        """
        out_set = self.outbound_edges(node_obj)
        if edge_set is None:
            return len(out_set)
        else:
            for e in edge_set:
                if e not in out_set:
                    return False
        return True

    def is_attached_in(self, node_obj, edge_set=None):
        """
        Tests whether edges attach from a node
        """
        in_set = self.inbound_edges(node_obj)
        if edge_set is None:
            return len(in_set)
        else:
            for e in edge_set:
                if e not in in_set:
                    return False
        return True

    def is_attached_bi(self, node_obj, edge_set=None):
        """
        Tests whether edges attach either to or from a node
        """
        bi_set = self.bidirected_edges(node_obj)
        if edge_set is None:
            return len(bi_set)
        else:
            for e in edge_set:
                if e not in bi_set:
                    return False
        return True

    def outbound_edges(self, node_obj=None):
        """Returns list of all edges out of node"""
        edge_set = []
        for e in self.edge_ids.obj_list():
            if not e.is_bidirected() and e.connects(self, node_obj):
                edge_set.append(e)
        return edge_set

    def inbound_edges(self, node_obj=None):
        """Returns list of all edges in to node"""
        edge_set = []
        for e in self.edge_ids.obj_list():
            if not e.is_bidirected() and e.connects(node_obj, self):
                edge_set.append(e)
        return edge_set

    def bidirected_edges(self, node_obj=None):
        """Returns list of all edges between nodes"""
        edge_set = []
        for e in self.edge_ids.obj_list():
            if e.is_bidirected() and e.connects(self, node_obj):
                edge_set.append(e)
        return edge_set

    def attached_edges(self, node_obj=None, silent=False):
        """
        Returns the edges connecting self to or from a node

        silent is used to indicate whther the edge is intened to be used.
        (effectively turns off error checking when false)
        """
        edge_set = (
            self.outbound_edges(node_obj)
            + self.inbound_edges(node_obj)
            + self.bidirected_edges(node_obj)
        )
        if not silent and edge_set is None:
            logger.error(str(f"No edges found between {self} and {node_obj}"))
        return edge_set

    def outbound_nodes(self, edge_set=None):
        """Returns list of all nodes connected by edges out"""
        if edge_set is None:
            edge_set = self.edge_ids.obj_list()
        ret_list = []
        for e in edge_set:
            if not e.is_bidirected() and e.connects(source=self):
                ret_list.append(e.to_node())
        return ret_list

    def inbound_nodes(self, edge_set=None):
        """Returns list of all nodes connected by edges in"""
        if edge_set is None:
            edge_set = self.edge_ids.obj_list()
        ret_list = []
        for e in edge_set:
            if not e.is_bidirected() and e.connects(target=self):
                ret_list.append(e.from_node())
        return ret_list

    def bidirected_nodes(self, edge_set=None):
        """Returns list of all nodes connected by edges"""
        if edge_set is None:
            edge_set = self.edge_ids.obj_list()
        ret_list = []
        for e in edge_set:
            if e.is_bidirected():
                ret_list.append(e.opposing_node(self))
        return ret_list

    def attached_nodes(self):
        """Returns list of all nodes connected"""
        edge_set = self.edge_ids.obj_list()
        ret_list = []
        for e in edge_set:
            ret_list.append(e.opposing_node(self))
        return ret_list

    def dimension_matches(self, node_obj, silent=True):
        """Test if dimension matches another node"""
        matches = self.dimension == node_obj.dimension
        if not matches and not silent:
            logger.error(
                str(
                    "'{}' cant connect to '{}' - dim mismatch {}->{}".format(
                        self, node_obj, self.dimension, node_obj.dimension
                    )
                )
            )
        return matches

    def make_member_of(self, node_obj):
        """
        Adds node to higher dimension node and does relevant checks
        """
        # check if valid inclusion
        if node_obj.dimension != self.dimension + 1:
            logger.error(
                str(
                    "'{}' cant be member to '{}' - dimension mismatch {}->{}".format(
                        self, node_obj, self.dimension, node_obj.dimension
                    )
                )
            )
        # adds self to hdgd and hdgd to list of owners
        else:
            node_obj.member_node_ids.add_obj(self)
            self.parent_node_ids.add_obj(node_obj)

    def make_owner_of(self, node_obj):
        """
        Adds node to lower dimension node and does relevant checks
        """
        node_obj.make_member_of(self)

    def leave_memebership_of(self, node_obj):
        """Remove node from higher dimension node"""
        node_obj.member_node_ids.remove_obj(self)
        self.parent_node_ids.remove_obj(node_obj)

    def disown(self, node_obj):
        """Remove node from higher dimension node"""
        node_obj.leave_membership_of(self)

    def destroy(self):
        """
        Destroys self from memory and persistent storage
        """
        for i in self.edge_ids.obj_list():
            i.destroy()
        super().destroy()

    def dot_str(self, node_map=None, detailed=False):
        """
        DOT representation
        """

        def handle_str(str):
            return str[:32].replace('"', '\\"')

        if node_map is None:
            nid = f"{uuid.UUID(self.jid).hex}"
        else:
            nid = f"{node_map.index(self.jid)}"

        dstr = f'"n{nid}" [ '

        if detailed:
            dstr += f'id="{uuid.UUID(self.jid).hex}", '

        dstr += f'label="n{nid}:{self.name}" '

        node_dict = self.context
        for i in self.private_values():
            node_dict.pop(i)

        if node_dict and detailed:
            for k, v in node_dict.items():
                if not isinstance(v, str) or v == "":
                    continue
                dstr += f', {k}="{handle_str(v)}"'

        dstr += " ]"

        return dstr + "\n"



"""Built in actions for Jaseci"""
from jaseci.actions.live_actions import jaseci_action
import os
import json


@jaseci_action()
def load_str(fn: str, max_chars: int = None):
    """Standard built in for loading from file to string"""
    with open(fn, "r") as file:
        data = file.read(max_chars)
    return data


@jaseci_action()
def load_json(fn: str):
    """Standard built in for loading json from file to dictionary"""
    with open(fn, "r") as file:
        data = json.load(file)
    return data


@jaseci_action()
def dump_str(fn: str, s: str):
    """Standard built in for dumping to file from string"""
    with open(fn, "w") as file:
        num_chars = file.write(s)
    return num_chars


@jaseci_action()
def append_str(fn: str, s: str):
    """Standard built in for appending to file from string"""
    with open(fn, "a") as file:
        num_chars = file.write(s)
    return num_chars


@jaseci_action()
def dump_json(fn: str, obj, indent: int = None):
    """Standard built in for dumping json to file from dictionary"""
    with open(fn, "w") as file:
        json.dump(obj, file, indent=indent)


@jaseci_action()
def delete(fn: str):
    """Standard built in for deleting a file"""
    if os.path.exists(fn):
        os.remove(fn)
        return True
    else:
        return False



prog0 = """
    node testnode:0 {
        has a, b, c;
        can std.log::a,b::>c with exit;
    }

    walker testwalk {
        testnode {
            here.a = 43;
            here.b = 'Yeah \\n"fools"!';
            report here.b;
            if(4 > 6) { std.log("a"); }
            elif(5>6) { std.log("b"); }
            elif(6>6) { std.log("c"); }
            elif(7>6) { std.log(576); }
        }
    }

    node life:0 {
    }

    node year {
        has anchor year;

    }

    walker another {
        life {
            here.a = 43;
            here.b = 'Yeah \\n"fools"!';
            report here.b;
            if("4 > 6" == "4 > 6") { std.log("a"); }
        }
    }
    """

prog1 = """
    node testnode:0 {
        has a, b, c;
        can std.log::a,b::>c with exit;
    }

    walker testwalk {
        testnode {
            here.a = 43;
            here.b = 'Yeah \\n"fools"!';
            report here.b;
            if(4 > 6) { std.log("a"); }
            elif(5>6) { std.log("b"); }
            elif(6>6) { std.log("c"); }
            elif(7>6) { std.log(576); }
        }
    }

    node life:0 {
    }

    node year {
        has anchor year;

    }

    node month {
        has anchor month;
    }

    node week {
        has anchor week;
    }

    node day {
        has anchor day;
    }

    node workette {
        has date, owner, status, snooze_till;
        has note, is_MIT, is_ritual;
    }

    walker use_test {
        can use.enc_question, use.enc_answer, use.qa_score;
        has output;
        q = use.enc_question(["How old are you?",
                              "which animal is the best?"]);
        std.log(q);
        a = use.enc_answer(["I'm 40 years old.", "Elephants rule."]);
        std.log(a);
        output = use.qa_score(q, a);
        report output;
    }

    walker use_test_with_ctx {
        can use.enc_question, use.enc_answer, use.qa_score, use.dist_score;
        has output;
        q = use.enc_question("Who are you?");
        a = use.enc_answer("I am jason");
        output = use.qa_score(q, a);
        report output;
        a = use.enc_answer("You are jon");
        output = use.qa_score(q, a);
        report output;
        a = use.enc_answer("Who are you? You are jon");
        output = use.qa_score(q, a);
        report output;
        a = use.enc_answer("Who are you? You are jon");
        output = use.qa_score(q, a);
        report output;
        q1 = use.enc_question("Who are you?");
        q2 = use.enc_question("Who you be?");
        q3 = use.enc_question("Who I be?");
        output = use.dist_score(q1, q2);
        report output;
        output = use.dist_score(q1, q3);
        report output;
        output = use.qa_score(q2, use.enc_answer("Who are you? You are jon"));
        report output;
        output = use.qa_score(q3, use.enc_answer("Who are you? You are jon"));
        report output;
        output = use.qa_score(q2, use.enc_answer("I am jason"));
        report output;
        output = use.qa_score(q3, use.enc_answer("I am jason"));
        report output;
    }

    walker use_test_with_ctx2 {
        can use.enc_question, use.enc_answer, use.qa_score, use.dist_score;

        q1 = use.enc_question("Who are you?");
        q2 = use.enc_question("Who you be?");
        q3 = use.enc_question("Who I be?");
        report use.dist_score(q1, q2);
        report use.dist_score(q1, q3);
        report use.qa_score(q2, use.enc_answer("Who are you? You are jon"));
        report use.qa_score(q3, use.enc_answer("Who are you? You are jon"));
        report use.qa_score(q2, use.enc_answer("I am jason"));
        report use.qa_score(q3, use.enc_answer("I am jason"));
        report use.qa_score(q3, use.enc_answer("I am jason","Who I be?"));
        report use.qa_score(q3, use.enc_answer("I am jason Who I be?"));
    }

    walker use_test_single {
        can use.enc_question, use.enc_answer, use.qa_score;
        has output;
        q = use.enc_question("Who's your daddy?");
        a = use.enc_answer("I'm your father.");
        output = use.qa_score(q, a);
        report output;
    }

    walker get_day {
        has date;
        life: take infer.year_from_date(date);
        year: take infer.month_from_date(date);
        month: take infer.week_from_date(date);
        week: take infer.day_from_date(date);
        day: report --> ;
    }

    walker get_gen_day {
        has date;
        can infer.year_from_date;
        can infer.month_from_date;
        can infer.week_from_date;
        can infer.day_from_date;
        life: take --> node::year == infer.year_from_date(date) else {
                new = spawn here --> node::year;
                new.year = infer.year_from_date(date);
                take --> node::year == infer.year_from_date(date);
            }
        year: take --> node::month == infer.month_from_date(date) else {
                new = spawn here --> node::month;
                new.month = infer.month_from_date(date);
                take --> node::month == infer.month_from_date(date);
            }
        month: take --> node::week == infer.week_from_date(date) else {
                new = spawn here --> node::week;
                new.week = infer.week_from_date(date);
                take --> node::week == infer.week_from_date(date);
            }
        week: take --> node::day == infer.day_from_date(date) else {
                new = spawn here --> node::day;
                new.day = infer.day_from_date(date);
                take --> node::day == infer.day_from_date(date);
            }
        day: report --> ;
    }

    walker get_sub_workettes {
        workette: report --> node::workette;
    }

    walker get_latest_day {
        life: take year.max_outbound;
        year: take month.max_outbound;
        month: take week.max_outbound;
        week: report day.max_outbound;
    }

    walker carry_forward {
        has my_root;
        day {
            new_day = spawn here --> node::day;
            my_root = new_day;
            take day.outbound_nodes;
        }
        workette {
            if(workette.status == 'done' or
            workette.status == 'eliminated') {
                continue;
            }
            childern = workette.outbound_nodes;
            new_workette = spawn here --> node::workette;
            parent = me.spawn_history.last(-1);
            new_workette <-- parent;
            take --> node::workette;
        }
        report me.spawn_history;
        report new_day;
    }
    """

edgey = """
    node testnode;

    edge apple;
    edge banana;

    walker init {
        root {
            a = spawn here --> node::testnode ;
            here -[apple]-> a;
            here -[banana]-> a;
        }
    }
    """

edgey2 = """
    node testnode;

    edge apple;
    edge banana;

    walker init {
        root {
            a = spawn here --> node::testnode ;
            here -[apple]-> a;
            here -[banana]-> a;

            here !--> a;
        }
    }
    """

edgey2b = """
    node testnode;

    edge apple;
    edge banana;

    walker init {
        root {
            a = spawn here --> node::testnode ;
            b = spawn here --> node::testnode ;
            here -[apple]-> a;
            here -[banana]-> a;

            here !--> a;
        }
    }
    """

edgey2c = """
    node testnode;

    edge apple;
    edge banana;

    walker init {
        root {
            a = spawn here --> node::testnode ;
            b = spawn here --> node::testnode ;
            here -[apple]-> a;
            here -[banana]-> a;

            here !-[apple]-> a;
        }
    }
    """

edgey3 = """
    node testnode;

    edge apple;
    edge banana;

    walker init {
        root {
            a = spawn here --> node::testnode ;
            b = spawn here --> node::testnode ;
            here -[apple]-> a;
            here -[apple]-> a;
            here -[banana]-> a;
            here -[banana]-> a;
            here -[banana]-> a;

            here !-[apple]-> a;
        }
    }
    """


edgey4 = """
    node testnode;

    edge apple;
    edge banana;

    walker init {
        root {
            a = spawn here --> node::testnode ;
            here --> a;
            here -[apple]-> a;
            here -[banana]-> a;

            here !-[generic]-> a;
        }
    }
    """

edgey5 = """
    node testnode;

    edge apple;
    edge banana;

    walker init {
        root {
            a = spawn here --> node::testnode ;
            here --> a;
            here --> a;
            here -[apple]-> a;
            here -[banana]-> a;

            here !-[generic]-> -[generic]->;
        }
    }
    """

edgey6 = """
    node testnode;

    edge apple;
    edge banana;

    walker init {
        root {
            a = spawn here --> node::testnode ;
            b = spawn here --> node::testnode ;

            here -[apple]-> -[generic]->;
        }
    }
    """

edgey7 = """
    node testnode;

    edge apple;
    edge banana;

    walker init {
        root {
            a = spawn here --> node::testnode ;
            b = spawn here --> node::testnode ;
            here --> a;
            here --> a;
            here -[apple]-> a;
            here -[apple]-> b;
            here -[banana]-> a;

            here !-[generic]-> -[apple]->;
        }
    }
    """

edge_access = """
    node testnode;

    edge apple {
        has v1, v2;
    }

    edge banana {
        has x1, x2;
    }

    walker init {
        root {
            a = spawn here -[apple]-> node::testnode ;
            b = spawn here -[banana]-> node::testnode ;

            e = -[apple]->.edge[0];
            e.v1 = 7;
            e = (--> node::testnode).edge[1];
            e.x1=8;
        }
    }
    """

has_assign = """
    node testnode {
        has a=8;
    }


    walker init {
        root {
            a = spawn here --> node::testnode ;
            b = spawn here --> node::testnode ;

            std.log(a.a, b.a);
        }
    }
    """


set_get_global = """
    walker setter {
        root {
            std.set_global('globby', 59);
        }
    }

    walker getter {
        has a;
        root {
            a=std.get_global('globby');
            std.log(std.get_global('globby'));
        }
    }
    """

set_get_global2 = """
    walker setter {
        root {
            std.set_global('globby2', 59);
        }
    }

    walker getter {
        has a;
        root {
            a=std.get_global('globby2');
            std.log(std.get_global('globby2'));
        }
    }
    """

set_get_global_dict = """
    walker setter {
        root {
            std.set_global('globby',
            { "max_bot_count": 10, "max_ans_count": 100,
              "max_txn_count": 50000, "max_test_suite": 5,
              "max_test_cases": 50, "export_import": true,
              "analytics": true, "integration": "All"
            });
        }
    }

    walker getter {
        has a;
        root {
            a=std.get_global('globby');
            std.log(std.get_global('globby'));
            report std.get_global('globby');
        }
    }
    """

version_label = """
    version: "alpha-1.0"

    walker setter {
        root {
            std.set_global('globby', 59);
        }
    }

    walker getter {
        has a;
        root {
            a=std.get_global('globby');
            std.log(std.get_global('globby'));
        }
    }
    """

sharable = """
    node life {
    }

    walker init {
        root {
            new = spawn here --> node::life;
            take -->;
        }
        life {
            std.out(here);
        }
    }
    """

basic = """
    node life {
    }

    walker init {
        root {
            new = spawn here --> node::life;
            take -->;
        }
        life {
        }
    }
    """

visibility_builtins = """
    node testnode {
        has yo, mama;
    }

    edge apple {
        has v1, v2;
    }

    edge banana {
        has x1, x2;
    }

    walker init {
        root {
            a = spawn here -[apple]-> node::testnode ;
            a.yo="Yeah i said";
            a.mama="Yo Mama Fool!";
            b = spawn here -[banana]-> node::testnode ;

            e = -[apple]->.edge[0];
            e.v1 = 7;
            e = --> node::testnode .edge[1];
            e.x1=8;

            report [a.context, b.info, e.details];
        }
    }
    """

spawn_ctx_edge_node = """
    node person: has name, age, birthday, profession;
    edge friend: has meeting_place;
    edge family: has kind;

    walker init {
        person1 = spawn here -[friend(meeting_place = "college")]->
            node::person(name = "Josh", age = 32);
        person2 = spawn here -[family(kind = "sister")] ->
            node::person(name = "Jane", age = 30);

        for i in -->{
            report i.context;
            report i.edge[0].context;
        }
    }
    """

filter_ctx_edge_node = """
    node person: has name, age, birthday, profession;
    edge friend: has meeting_place;
    edge family: has kind;

    walker init {
        person1 = spawn here -[friend(meeting_place = "college")]->
            node::person(name = "Josh", age = 32);
        person2 = spawn here -[family(kind = "sister")] ->
            node::person(name = "Jane", age = 30);

        report --> node::person(name=='Jane')[0].context;
        report -[family(kind=="brother")]->;
    }
    """

null_handleing = """
    node person: has name, age, birthday, profession;

    walker init {
        person1 = spawn here -->
            node::person(name = "Josh", age = 32);

        if(person1.birthday==null): report true;
        else: report false;

        if(person1.name==null): report true;
        else: report false;

        person1.name=null;
        report person1.name==null;
        person1.name=0;
        report person1.name==null;
    }
    """

bool_type_convert = """
    node person: has name;

    walker init {
        p1 = spawn here -->
            node::person(name = "Josh");

        p1.name = true;
        report p1.name;
        std.log(p1.name);
        report p1.context;
    }
    """

typecasts = """
    walker init {
        a=5.6;
        report (a+2);
        report (a+2).int;
        report (a+2).str;
        report (a+2).bool;
        report (a+2).int.float;

        if(a.str.type == str and !(a.int.type == str)
           and a.int.type == int):
            report "Types comes back correct";
    }
    """

typecasts_error = """
    walker init {
        a=5.6;
        report (a+2);
        report (a+2).int;
        report (a+2).str;
        report (a+2).edge;
        report ("a+2").int.float;

        if(a.str.type == str and !(a.int.type == str)
           and a.int.type == int):
            report "Types comes back correct";
    }
    """

filter_on_context = """
    node testnode {
        has yo, mama;
    }

    edge apple {
        has v1, v2;
    }

    edge banana {
        has x1, x2;
    }

    walker init {
        root {
            a = spawn here -[apple]-> node::testnode ;
            a.yo="Yeah i said";
            a.mama="Yo Mama Fool!";
            b = spawn here -[banana]-> node::testnode ;

            e = -[apple]->.edge[0];
            e.v1 = 7;
            e = --> node::testnode .edge[1];
            e.x1=8;

            report [a.context.{yo}, b.info.{jid,j_type}, e.details];
        }
    }
    """

string_manipulation = """
    walker init {
        a=" tEsting me  ";
        report a[4];
        report a[4:7];
        report a[3:-1];
        report a.str::upper;
        report a.str::lower;
        report a.str::title;
        report a.str::capitalize;
        report a.str::swap_case;
        report a.str::is_alnum;
        report a.str::is_alpha;
        report a.str::is_digit;
        report a.str::is_title;
        report a.str::is_upper;
        report a.str::is_lower;
        report a.str::is_space;
        report '{"a": 5}'.str::load_json;
        report a.str::count('t');
        report a.str::find('i');
        report a.s::split;
        report a.s::split('E');
        report a.s::startswith('tEs');
        report a.str::endswith('me');
        report a.str::replace('me', 'you');
        report a.str::strip;
        report a.str::strip(' t');
        report a.str::lstrip;
        report a.str::lstrip(' tE');
        report a.str::rstrip;
        report a.str::rstrip(' e');

        report a.str::upper.str::is_upper;
    }
    """

list_manipulation = """
    walker init {
        a = [4];
        b=a.l::copy;
        b[0]+=1;
        report a;
        report b;
        a.list::extend(b);
        a.l::append(b[0]);
        report a;
        a.l::reverse;
        report a;
        a.list::sort;
        report a;
        a.l::reverse;
        report a.l::index(4);
        a.l::append(a.l::index(4));
        report a;
        a.l::insert(2, "apple");
        a.l::remove(5);
        report a;
        a.l::pop;
        report a.l::count(4);
        report a;
        a.l::clear;
        report a;
    }
    """

list_reversed = """
    walker init {
        a = [4, 2, 7];
        report a.l::reversed;
        report a;
    }
    """

dict_manipulation = """
    walker init {
        a = {'four':4, 'five':5};
        b=a.d::copy;
        b['four']+=1;
        report a;
        report b;
        report a.dict::items;
        report a.d::keys;
        a.d::popitem;
        report a;
        report a.dict::values;
        a.d::update({'four': 7});
        report a;
        a.d::pop('four');
        report a;
    }
    """

string_join = """
    walker init {
        a=['test', 'me', 'now'];
        report '_'.str::join(a);
    }
    """

sub_list = """
    walker init {
        a=[1,2,3,4,5,6,7,8,9];
        report a[4:7];
    }
    """

destroy_and_misc = """
    node person: has name, age, birthday, profession;
    edge friend: has meeting_place;
    edge family: has kind;

    walker init {
        person1 = spawn here -[friend(meeting_place = "college")]->
            node::person(name = "Josh", age = 32);
        person2 = spawn here -[family(kind = "sister")] ->
            node::person(name = "Jane", age = 30);

        report person1.name;
        destroy person1.name;
        report person1.context;
        person1.name="pete";
        report person1.context;
        a=[1,2,3];
        destroy a[1];
        report a;
        b={'a': 'b', 'c':'d'};
        destroy b['c'];
        report b;
        a=[1,2,3,5,6,7,8,9];
        destroy a[2:4];
        report a;
        a[2:4]=[45,33];
        report a;
        destroy a;
        report a;
        person1.banana=45;
        report person1.context;
        report 'age' in person1.context;
    }
    """

arbitrary_assign_on_element = """
    node person: has name, age, birthday, profession;
    walker init {
        some = spawn here --> node::person;
        some.apple = 45;
        report some.context;
    }
    """

try_else_stmts = """
    walker init {
        a=null;
        try {a=2/0;}
        else with err {report err;}
        try {a=2/0;}
        else {report 'dont need err';}
        try {a=2/0;}
        try {a=2/0;}
        report a;
        try {a=2/1;}
        report a;
    }
    """

node_edge_same_name = """
    node person: has name, age, birthday, profession;
    edge person: has meeting_place;

    walker init {
        person1 = spawn here -[person(meeting_place = "college")]->
            node::person(name = "Josh", age = 32);

        report -->.edge[0].context;
        report -->[0].context;
    }
    """

testcases = """
    node testnode {
        has yo, mama;
    }

    node apple {
        has v1, v2;
    }

    node banana {
        has x1, x2;
    }

    graph dummy {
        has anchor graph_root;
        spawn {
            graph_root = spawn node::testnode (yo="Hey yo!");
            n1=spawn node::apple(v1="I'm apple");
            n2=spawn node::banana(x1="I'm banana");
            graph_root --> n1 --> n2;
        }
    }

    walker init {
        has num=4;
        report here.context;
        report num;
        take -->;
    }

    test "basic test with refs"
    with graph::dummy by walker::init;

    test "test with refs and assert block"
    with graph::dummy by walker::init {
       report "ASSERT BLOCK";
    }

    test "test with graph ref and walker block"
    with graph::dummy by walker {
        report here.context;
        report "IN generic walker";
        take -->;
    }

    test "test with graph block and walker ref"
    with graph {
        has anchor graph_root;
        spawn {
            graph_root = spawn node::testnode (yo="Hey yo!");
            n1=spawn node::apple(v1="I'm apple");
            n2=spawn node::banana(x1="I'm banana");
            graph_root --> n1 --> n2;
            graph_root --> n2;
        }
    } by walker::init {
        report "ASSERT BLOCK";
    }
    """


testcase_asserts = """
    node testnode {
        has yo, mama;
    }

    node apple {
        has v1, v2;
    }

    node banana {
        has x1, x2;
    }

    graph dummy {
        has anchor graph_root;
        spawn {
            graph_root = spawn node::testnode (yo="Hey yo!");
            n1=spawn node::apple(v1="I'm apple");
            n2=spawn node::banana(x1="I'm banana");
            graph_root --> n1 --> n2;
        }
    }

    walker init {
        has num=4;
        report here.context;
        report num;
        take -->;
    }

    test "assert should be valid"
    with graph::dummy by walker::init {
       assert (num==4);
       assert (here.x1=="I'm banana");
       assert <--[0].v1=="I'm apple";
    }

    test "assert should fail"
    with graph::dummy by walker::init {
       assert (num==4);
       assert (here.x1=="I'm banana");
       assert <--[0].v1=="I'm Apple";
    }

    test "assert should fail, add internal except"
    with graph::dummy by walker::init {
       assert (num==4);
       assert (here.x1=="I'm banana");
       assert <--[10].v1=="I'm apple";
    }
    """

report_not_to_jacset = """
    node testnode {
        has yo, mama;
    }

    walker init {
        spawn here --> node::testnode;
        report -->;
    }
    """

walker_spawn_unwrap_check = """
    node testnode {
        has yo, mama;
    }

    walker print {
       has anchor nd;
       nd=here;
    }

    walker init {
        report &(spawn here walker::print);
    }
    """

std_get_report = """
    walker init {
       report 3;
       report 5;
       report 6;
       report 7;
       report std.get_report();
       report 8;
    }
    """

func_with_array_index = """
    walker init {
       report 3;
       report 5;
       report std.get_report()[0];
    }
    """

rt_error_test1 = """
    walker init {
       spawn here --> node::generic;
       report -->[2];
    }
    """


root_type_nodes = """
    walker init {
       spawn here -[generic]-> node::root;
       report here.details['name'];
       report -->[0].details['name'];
    }
    """

invalid_key_error = """
    walker init {
       report here.context['adfas'];
    }
    """

auto_cast = """
    walker init {
        report 1==1.0;
        report 1.0==1;
    }
    """

no_error_on_dict_key_assign = """
    walker init {
        a={};
        a['b']=4;
        report a;
    }
    """

report_status = """
    walker init {report:status = 302; report "hello";}
    """


graph_in_graph = """
    graph one {
        has anchor graph_root;
        spawn {
            graph_root = spawn node::generic;
        }
    }

    graph two {
        has anchor graph_root;
        spawn {

            graph_root = spawn node::generic;
            day1 = spawn graph::one;

            graph_root --> day1;
        }
    }

    walker init {
        root {
            spawn here --> graph::two;
        }
        take -->;
        report here;
    }
    """


min_max_on_list = """
    walker init {
        a = [45, 3, 531.0, 3, 6, 531.1];

        report a.l::max;
        report a.l::min;
        report a.l::idx_of_max;
        report a.l::idx_of_min;
    }
    """

edge_bug = """
    node plain;

    edge g;

    walker init {
        root {
            nd = spawn here -[g]-> node::plain;
            nd -[g]-> nd;
            spawn nd -[g]-> node::plain;
            spawn nd -[g]-> node::plain;
            a = spawn nd <-[g]- node::plain;
            spawn nd <-[g]- node::plain;
            nd -[g]-> a;
        }
        take -[g]->;
        plain {
            report -[g]->.edge;
            disengage;
        }
    }
    """

rand_choice = """
    walker init {
        a = [45, 3, 531.0, 3, 6, 531.1];

        report a;
        report rand.choice(a);
    }
    """


struct_types = """
    type simple {
        has apple;
        has orange=43;
    }

    type another {
        has apple=type::simple(apple=33);
    }

    node mynode {
        has basic=type::another;
    }

    walker init {
        a = spawn here --> node::mynode;

        report
        [a.basic["apple"]["orange"],
         a.basic["apple"]["apple"]];
        report [a.basic.apple.apple, a.basic.apple.orange];
    }
    """



import re
from copy import deepcopy
from multiprocessing import Process
from typing import Tuple
from uuid import UUID

from celery import Task
from celery.app.control import Inspect
from requests import get, post
from requests.exceptions import HTTPError

DEFAULT_MSG = "Skipping scheduled walker!"


class Queue(Task):
    def run(self, wlk, nd, args):
        from jaseci.svc import MetaService

        hook = MetaService().build_hook()

        wlk = hook.get_obj_from_store(UUID(wlk))
        wlk._to_await = True

        nd = hook.get_obj_from_store(UUID(nd))
        resp = wlk.run(nd, *args)
        wlk.destroy()

        return {"anchor": wlk.anchor_value(), "response": resp}


class ScheduledWalker(Task):
    def get_obj(self, jid):
        return self.hook.get_obj_from_store(UUID(jid))

    def run(self, name, ctx, nd=None, snt=None, mst=None):
        from jaseci.svc import MetaService

        self.hook = MetaService().build_hook()

        if mst:
            mst = self.get_obj(mst)
        else:
            return f"{DEFAULT_MSG} mst (Master) is required!"

        if mst is None:
            return f"{DEFAULT_MSG} Invalid Master!"

        try:
            if not snt:
                if mst.active_snt_id == "global":
                    global_snt_id = self.hook.get_glob("GLOB_SENTINEL")
                    snt = self.get_obj(global_snt_id)
                elif mst.active_snt_id:
                    snt = self.get_obj(mst.active_snt_id)
            elif snt in mst.alias_map:
                snt = self.get_obj(mst.alias_map[snt])
            else:
                snt = self.get_obj(snt)

            if not snt:
                return f"{DEFAULT_MSG} Invalid Sentinel!"

            if not nd:
                if mst.active_gph_id:
                    nd = self.get_obj(mst.active_gph_id)
            elif nd in mst.alias_map:
                nd = self.get_obj(mst.alias_map[nd])
            else:
                nd = self.get_obj(nd)

            if not nd:
                return f"{DEFAULT_MSG} Invalid Node!"

            return mst.walker_run(name, nd, ctx, ctx, snt, False, False)
        except Exception as e:
            return f"{DEFAULT_MSG} Error occured: {e}"


class ScheduledSequence(Task):
    json_escape = re.compile(r"[^a-zA-Z0-9_]")
    internal = re.compile(r"\(([a-zA-Z0-9_\.\[\]\$\#\@\!]*?)\)")
    full = re.compile(r"^\{\{([a-zA-Z0-9_\.\[\]\$\#\(\)\@\!]*?)\}\}$")
    partial = re.compile(r"\{\{([a-zA-Z0-9_\.\[\]\$\#\(\)\@\!]*?)\}\}")

    def get_deep_value(self, data, keys, default):
        if len(keys) == 0:
            return data

        key = keys.pop(0)
        t = type(data)

        if t is dict and key in data:
            return self.get_deep_value(data[key], keys, default)
        elif t is list and key.isnumeric():
            return self.get_deep_value(data[int(key)], keys, default)
        else:
            return default

    def get_value(self, holder: Tuple, keys: str, default=None):
        while self.internal.search(keys):
            for intern in self.internal.findall(keys):
                keys = keys.replace(
                    "(" + intern + ")", self.get_value(holder, intern, "")
                )

        if keys:
            keys = keys.split(".")
            key = keys.pop(0)
            if key == "#":
                return self.get_deep_value(holder[0], keys, default)
            elif key == "$":
                t = type(holder[1])
                if t is dict or t is list:
                    return self.get_deep_value(holder[1], keys, default)
                else:
                    return holder[1]
            elif key == "@":
                t = type(holder[2])
                if t is dict or t is list:
                    return self.get_deep_value(holder[2], keys, default)
                else:
                    return holder[2]
            elif key == "!":
                return holder[3]
        return default

    def compare(self, condition, expected, actual):
        if condition == "eq":
            return actual == expected
        elif condition == "ne":
            return actual != expected
        elif condition == "gt":
            return actual > expected
        elif condition == "gte":
            return actual >= expected
        elif condition == "lt":
            return actual < expected
        elif condition == "lte":
            return actual <= expected
        elif condition == "regex":
            return re.compile(expected).match(actual)

    def condition(self, holder: Tuple, filter):
        for cons in filter["condition"].keys():
            if not (filter["condition"][cons] is None) and not self.compare(
                cons, filter["condition"][cons], self.get_value(holder, filter["by"])
            ):
                return False
        return True

    def or_condition(self, holder: Tuple, filter):
        for filt in filter:
            if "condition" in filt and self.condition(holder, filt):
                return True
            elif "or" in filt and self.or_condition(holder, filt["or"]):
                return True
            elif "and" in filt and self.and_condition(holder, filt["and"]):
                return True
        return False

    def and_condition(self, holder: Tuple, filter):
        for filt in filter:
            if "condition" in filt and not self.condition(holder, filt):
                return False
            elif "or" in filt and not self.or_condition(holder, filt["or"]):
                return False
            elif "and" in filt and not self.and_condition(holder, filt["and"]):
                return False
        return True

    def deep_replace_str(self, holder: Tuple, data, key):
        matcher = self.full.match(data[key])
        if matcher:
            data[key] = self.get_value(holder, matcher.group(1))
        else:
            for rep in self.partial.findall(data[key]):
                data[key] = data[key].replace(
                    "{{" + rep + "}}", self.get_value(holder, rep, "")
                )

    def deep_replace_dict(self, holder: Tuple, data):
        for key in data.keys():
            if key != "__def_loop__":
                t = type(data[key])
                if t is str:
                    self.deep_replace_str(holder, data, key)
                elif t is dict:
                    self.deep_replace_dict(holder, data[key])

    def save(self, holder: Tuple, req, params):
        if params in req:
            holder[0][self.json_escape.sub("_", req[params])] = holder[1]

    def trigger_interface(self, req: dict):
        from jaseci.svc import MetaService

        master = req.get("master")
        app = MetaService()
        if master is None:
            caller = app.build_master()
            trigger_type = "public"
        else:
            caller = app.build_hook().get_obj_from_store(master)
            trigger_type = "general"

        api = req.get("api")
        body = req.get("body", {})

        return getattr(caller, f"{trigger_type}_interface_to_api")(body, api)

    def run(self, *args, **kwargs):
        requests = kwargs.get("requests")
        persistence = kwargs.get("persistence", {})
        container = kwargs.get("container", {"current": persistence})
        index = container.get("index", "0")

        if "parent_current" in container:
            container["current"] = container["parent_current"]

        for req in requests:
            try:
                self.deep_replace_dict(
                    (
                        persistence,
                        container["current"],
                        container.get("parent_current", {}),
                        index,
                    ),
                    req,
                )
                self.save((persistence, req), req, "save_req_to")

                method = req["method"].upper()

                if method == "JAC":
                    container["current"] = self.trigger_interface(req)
                else:
                    if method == "POST":
                        response = post(
                            req["api"],
                            json=req.get("body", {}),
                            headers=req.get("header", {}),
                        )
                    elif method == "GET":
                        response = get(req["api"], headers=req.get("header", {}))
                    response.raise_for_status()
                    if "application/json" in response.headers.get("Content-Type"):
                        container["current"] = response.json()
                    else:
                        container["current"] = response.text

                if "__def_loop__" in req:
                    def_loop = req["__def_loop__"]
                    for idx, loop in enumerate(
                        self.get_value(
                            (persistence, container["current"]), def_loop["by"], []
                        )
                    ):
                        if "filter" in def_loop and not self.and_condition(
                            (persistence, loop), def_loop["filter"]
                        ):
                            continue
                        loop_container = {"parent_current": loop, "index": str(idx)}
                        self.run(
                            requests=deepcopy(def_loop["requests"]),
                            persistence=persistence,
                            container=loop_container,
                        )

                self.save((persistence, container["current"]), req, "save_to")

            except Exception as err:
                container["current"] = (
                    {
                        "status": err.response.status_code,
                        "message": err.response.reason,
                    }
                    if isinstance(err, HTTPError)
                    else {
                        "worker_error": str(err),
                    }
                )

                self.save((persistence, container["current"]), req, "save_to")

                if "ignore_error" not in req or not req["ignore_error"]:
                    if "parent_current" in container:
                        raise err
                    break

            if "break" in req and self.and_condition(
                (persistence, container["current"]), req["break"]
            ):
                break

        return persistence


c1 = Queue
c2 = ScheduledWalker
c3 = ScheduledSequence


class TaskProperties:
    def __init__(self, prop):
        if not hasattr(prop, "_inspect"):
            setattr(prop, "_inspect", None)
            setattr(prop, "_worker", None)
            setattr(prop, "_scheduler", None)

            # --------------- REGISTERED TASK --------------- #
            setattr(prop, "_queue", None)
            setattr(prop, "_scheduled_walker", None)
            setattr(prop, "_scheduled_sequence", None)

    @property
    def inspect(self) -> Inspect:
        return self.cls._inspect

    @inspect.setter
    def inspect(self, val: Inspect):
        self.cls._inspect = val

    @property
    def worker(self) -> Process:
        return self.cls._worker

    @worker.setter
    def worker(self, val: Process):
        self.cls._worker = val

    @property
    def scheduler(self) -> Process:
        return self.cls._scheduler

    @scheduler.setter
    def scheduler(self, val: Process):
        self.cls._scheduler = val

    @property
    def queue(self) -> c1:
        return self.cls._queue

    @queue.setter
    def queue(self, val: c1):
        self.cls._queue = val

    @property
    def scheduled_walker(self) -> c2:
        return self.cls._scheduled_walker

    @scheduled_walker.setter
    def scheduled_walker(self, val: c2):
        self.cls._scheduled_walker = val

    @property
    def scheduled_sequence(self) -> c3:
        return self.cls._scheduled_sequence

    @scheduled_sequence.setter
    def scheduled_sequence(self, val: c3):
        self.cls._scheduled_sequence = val



import signal
import sys
from multiprocessing import Process

from celery import Celery
from celery.backends.base import DisabledBackend

from jaseci.svc import CommonService, ServiceState as Ss
from jaseci.utils.utils import logger
from .common import Queue, ScheduledSequence, ScheduledWalker, TaskProperties

################################################
#                   DEFAULTS                   #
################################################

TASK_CONFIG = {
    "enabled": True,
    "quiet": True,
    "broker_url": "redis://localhost:6379/1",
    "result_backend": "redis://localhost:6379/1",
    "broker_connection_retry_on_startup": True,
    "task_track_started": True,
}


#################################################
#                   TASK APP                   #
#################################################


class TaskService(CommonService, TaskProperties):

    ###################################################
    #                   INITIALIZER                   #
    ###################################################

    def __init__(self, hook=None):
        CommonService.__init__(self, TaskService)
        TaskProperties.__init__(self, self.cls)

        try:
            if self.is_ready() and self.__has_redis(hook):
                self.state = Ss.STARTED
                self.__task(hook)
        except Exception as e:
            if not (self.quiet):
                logger.error(
                    "Skipping Celery due to initialization failure!\n"
                    f"{e.__class__.__name__}: {e}"
                )

            self.app = None
            self.state = Ss.FAILED
            self.terminate_worker()
            self.terminate_scheduler()

    def __task(self, hook):
        configs = self.get_config(hook)
        enabled = configs.pop("enabled", True)

        if enabled:
            self.quiet = configs.pop("quiet", False)
            self.app = Celery("celery")
            self.app.conf.update(**configs)
            self.inspect = self.app.control.inspect()
            self.__ping()
            self.__tasks()
            self.state = Ss.RUNNING

            self.__worker()
            self.__scheduler()
        else:
            self.state = Ss.DISABLED

    def __worker(self):
        self.worker = Process(target=self.app.Worker(quiet=self.quiet).start)
        self.worker.daemon = True
        self.worker.start()

    def __scheduler(self):
        self.scheduler = Process(target=self.app.Beat(quiet=self.quiet).run)
        self.scheduler.daemon = True
        self.scheduler.start()

    def __tasks(self):
        self.queue = self.app.register_task(Queue())
        self.scheduled_walker = self.app.register_task(ScheduledWalker())
        self.scheduled_sequence = self.app.register_task(ScheduledSequence())

    def __has_redis(self, hook=None):
        if not (hook is None) and hook.redis.has_failed():
            if not (self.quiet):
                logger.error(
                    "Redis is not yet running reason "
                    "for skipping Celery initialization!"
                )
            self.state = Ss.FAILED
            return False
        return True

    def __ping(self):  # will throw exception
        self.inspect.ping()
        self.app.AsyncResult("").result

    ###################################################
    #              COMMON GETTER/SETTER               #
    ###################################################

    def get_by_task_id(self, task_id, wait=False):
        task = self.app.AsyncResult(task_id)

        if isinstance(task.backend, DisabledBackend):
            return {
                "status": "DISABLED",
                "result": "result_backend is set to disabled!",
            }

        ret = {"status": task.state}
        if task.ready():
            ret["result"] = task.result
        elif wait:
            ret["status"] = "SUCCESS"
            ret["result"] = task.get(disable_sync_subtasks=False)

        return ret

    def inspect_tasks(self):
        return {
            "scheduled": self.inspect.scheduled(),
            "active": self.inspect.active(),
            "reserved": self.inspect.reserved(),
        }

    ###################################################
    #                     QUEUING                     #
    ###################################################

    def add_queue(self, wlk, nd, *args):
        return self.queue.delay(wlk.id.urn, nd.id.urn, args).task_id

    ###################################################
    #                     CLEANER                     #
    ###################################################

    def terminate_worker(self=None):
        if not (self.worker is None):
            if not (self.quiet):
                logger.warn("Terminating task worker ...")
            self.worker.terminate()
            self.worker = None

    def terminate_scheduler(self=None):
        if not (self.scheduler is None):
            if not (self.quiet):
                logger.warn("Terminating task scheduler ...")
            self.scheduler.terminate()
            self.scheduler = None

    def reset(self, hook):
        self.terminate_worker()
        self.terminate_scheduler()
        self.inspect = None
        self.build(hook)

    ####################################################
    #                    OVERRIDDEN                    #
    ####################################################

    def get_config(self, hook) -> dict:
        self.__scheduler = lambda: None
        return hook.build_config("TASK_CONFIG", TASK_CONFIG)


# ----------------------------------------------- #


###################################################
#                 PROCESS CLEANER                 #
###################################################


def terminate_gracefully(*args):
    sys.exit(0)


signal.signal(signal.SIGINT, terminate_gracefully)



"""
Master api as a mixin
"""
from jaseci.api.interface import Interface
from jaseci.utils.id_list import IdList
import uuid


class MasterApi:
    """
    Master APIs for users creating and managing groups of sub users

    These APIs
    """

    def __init__(self, head_master):
        self.caller = None
        self.head_master_id = head_master
        self.sub_master_ids = IdList(self)

    @Interface.private_api(cli_args=["name"])
    def master_create(
        self,
        name: str,
        global_init: str = "",
        global_init_ctx: dict = {},
        other_fields: dict = {},
    ):
        """
        Create a master instance and return root node master object

        other_fields used for additional feilds for overloaded interfaces
        (i.e., Dango interface)
        """
        if self.sub_master_ids.has_obj_by_name(name):
            return {"response": f"{name} already exists", "success": False}
        ret = {}
        mast = self.user_creator(name, other_fields)
        ret["user"] = mast.serialize()
        if len(global_init):
            ret["global_init"] = self.user_global_init(
                mast, global_init, global_init_ctx
            )
        self.take_ownership(mast)
        ret["success"] = True
        return ret

    @Interface.private_api(cli_args=["name"])
    def master_get(self, name: str, mode: str = "default", detailed: bool = False):
        """
        Return the content of the master with mode
        Valid modes: {default, }
        """
        mas = self.sub_master_ids.get_obj_by_name(name)
        if not mas:
            return {"response": f"{name} not found"}
        else:
            return mas.serialize(detailed=detailed)

    @Interface.private_api()
    def master_list(self, detailed: bool = False):
        """
        Provide complete list of all master objects (list of root node objects)
        """
        masts = []
        for i in self.sub_master_ids.obj_list():
            masts.append(i.serialize(detailed=detailed))
        return masts

    @Interface.private_api(cli_args=["name"])
    def master_active_set(self, name: str):
        """
        Sets the default master master should use
        NOTE: Specail handler included in general_interface_to_api
        """
        mas = self.sub_master_ids.get_obj_by_name(name)
        if not mas:
            return {"response": f"{name} not found"}
        self.caller = mas.jid
        return {"response": f"You are now {mas.name}"}

    @Interface.private_api()
    def master_active_unset(self):
        """
        Unsets the default sentinel master should use
        """
        self.caller = None
        return {"response": f"You are now {self.name}"}

    @Interface.private_api()
    def master_active_get(self, detailed: bool = False):
        """
        Returns the default master master is using
        """
        if self.caller:
            return self._h.get_obj(self._m_id, uuid.UUID(self.caller)).serialize(
                detailed=detailed
            )
        else:
            return self.serialize(detailed=detailed)

    @Interface.private_api()
    def master_self(self, detailed: bool = False):
        """
        Returns the masters object
        """
        return self.serialize(detailed=detailed)

    @Interface.private_api(cli_args=["name"])
    def master_delete(self, name: str):
        """
        Permanently delete master with given id
        """
        if not self.sub_master_ids.has_obj_by_name(name):
            return {"response": f"{name} not found", "success": False}
        self.sub_master_ids.destroy_obj_by_name(name)
        self.user_destroyer(name)
        return {"response": f"{name} has been destroyed", "success": True}

    def take_ownership(self, m):
        """
        Assumes owenership over another master
        """
        m.head_master_id = self.jid
        m.give_access(self)
        m.save()
        self.sub_master_ids.add_obj(m)

    def destroy(self):
        """
        Destroys self from memory and persistent storage
        """
        for i in self.sub_master_ids.obj_list():
            i.destroy()



from jaseci.svc import (
    CommonService,
    MetaProperties,
    MailService,
    RedisService,
    ServiceState as Ss,
    TaskService,
)

from jaseci.utils.utils import logger


class MetaService(CommonService, MetaProperties):
    def __init__(self, run_svcs=True):
        self.run_svcs = run_svcs
        CommonService.__init__(self, MetaService)
        MetaProperties.__init__(self, self.cls)

        if self.is_ready():
            self.state = Ss.STARTED

            self.build_classes()
            self.build_services()

            self.state = Ss.RUNNING

    ###################################################
    #                    SERVICES                     #
    ###################################################

    # args & kwargs will be used as *args & **kwargs on initialization
    def add_service_builder(self, name, svc, custom=False, args=[], kwargs={}):
        if self.services.get(name):
            raise Exception(f"{name} already exists!")

        if not custom and not issubclass(svc, CommonService):
            raise Exception(
                f"{svc.__class__.__name__} is not instance of CommonService!"
            )

        self.services[name] = {"cls": svc, "args": args, "kwargs": kwargs}

    def run_service(self, name, background=False, *args, **kwargs):

        svc = self.services.get(name, False)

        if not svc:
            logger.error(f"Service {name} is not yet set!")
            return None

        kwargs.update(svc["kwargs"])

        svc = svc["cls"](*args, *svc["args"], **kwargs)

        if background:
            self.background[name] = svc

        return svc

    def get_service(self, name, *args, **kwargs):
        svc = self.background.get(name, False)

        if not svc:
            return self.run_service(name, True, *args, **kwargs)

        return svc

    ###################################################
    #                     BUILDER                     #
    ###################################################

    def build_hook(self):
        params = self.hook_param
        h = self.hook(*params.get("args", []), **params.get("kwargs", []))
        if self.run_svcs:
            h.redis = self.get_service("redis", h)
            h.task = self.get_service("task", h)
            h.mail = self.get_service("mail", h)
            h.meta = self

        return h

    def build_master(self, *args, **kwargs):
        return self.__common(self.master, *args, **kwargs)

    def build_super_master(self, *args, **kwargs):
        return self.__common(self.super_master, *args, **kwargs)

    def __common(self, cls, *args, **kwargs):

        if not kwargs.get("h", None):
            kwargs["h"] = self.build_hook()

        return cls(*args, **kwargs)

    ###################################################
    #                   OVERRIDEN                     #
    ###################################################

    def build_classes(self):
        from jaseci.hook import RedisHook
        from jaseci.element.master import Master
        from jaseci.element.super_master import SuperMaster

        self.hook = RedisHook
        self.hook_param = {
            "args": [],  # override to add arg
            "kwargs": {},  # override to add field
        }
        self.master = Master
        self.super_master = SuperMaster

    def build_services(self):
        self.add_service_builder("redis", RedisService)
        self.add_service_builder("task", TaskService)
        self.add_service_builder("mail", MailService)



"""
Graph  class for Jaseci

"""
from collections import OrderedDict
from jaseci.graph.node import Node
from jaseci.utils.id_list import IdList


class Graph(Node):
    """Graph class for Jaseci"""

    def __init__(self, **kwargs):
        self.hd_node_ids = IdList(self)
        Node.__init__(self, **kwargs)
        self.name = "root"
        self.kind = "node"

        # Node.__init__(self, name="root", kind="node", **kwargs)

    def get_all_nodes(self, depth: int = 0):
        """
        Returns all reachable nodes
        """

        childs = {self.jid: self}
        nodes = OrderedDict(childs)
        depth -= 1

        while len(childs) and depth != 0:
            new_childs = OrderedDict()
            for child in childs.values():
                for _ch in child.attached_nodes():
                    if not (_ch.jid in nodes):
                        new_childs.update({_ch.jid: _ch})

            childs = new_childs
            nodes.update(childs)
            depth -= 1

        return nodes.values()

    def get_all_edges(self, nodes: list = None, depth: int = 0):
        """
        Returns all reachable edges
        """
        edges = OrderedDict()
        node_list = self.get_all_nodes(depth=depth) if nodes is None else nodes

        for nd in node_list:
            for _ch in nd.attached_edges():
                if not (_ch.jid in edges) and (
                    _ch.to_node() in node_list and _ch.from_node() in node_list
                ):
                    edges.update({_ch.jid: _ch})

        return edges.values()

    def graph_dot_str(self, detailed=False, depth: int = 0):
        """
        DOT representation for graph.
        NOTE: This is different from the dot_str method for node intentionally
        because graph inherits node.
        """
        node_list = self.get_all_nodes(depth=depth)
        edge_list = self.get_all_edges(nodes=node_list)
        node_map = [i.jid for i in node_list]
        edge_map = [i.jid for i in edge_list]

        # Construct the graph string
        dstr = ""
        dstr += f"strict digraph {self.name} {{\n"
        for n in node_list:
            dstr += f"    {n.dot_str(node_map, detailed)}"
        for e in edge_list:
            dstr += f"    {e.dot_str(node_map, edge_map, detailed)}"
        dstr += "}"
        return dstr

    def destroy(self):
        """
        Destroys self from memory and persistent storage
        """
        for i in self.hd_node_ids.obj_list():
            i.destroy()
        super().destroy()



"""
Walker api functions as a mixin
"""
from jaseci.api.interface import Interface
from jaseci.actor.walker import Walker
from jaseci.graph.node import Node
from jaseci.actor.sentinel import Sentinel
from jaseci.utils.id_list import IdList


class WalkerApi:
    """
    Walker APIs

    The walker set of APIs are used for execution and management of walkers. Walkers
    are the primary entry points for running Jac programs. The
    primary API used to run walkers is \\textbf{walker_run}. There are a number of
    variations on this API that enable the invocation of walkers with various
    semantics.
    """

    def __init__(self):
        self.spawned_walker_ids = IdList(self)
        self.yielded_walkers_ids = IdList(self)

    @Interface.private_api(cli_args=["wlk"])
    def walker_get(self, wlk: Walker, mode: str = "default", detailed: bool = False):
        """
        Get a walker rendered with specific mode
        Valid modes: {default, code, ir, keys, }
        """
        if mode == "code":
            return wlk._jac_ast.get_text()
        elif mode == "ir":
            return wlk.ir_dict()
        elif mode == "keys":
            return wlk.namespace_keys()
        else:
            return wlk.serialize(detailed=detailed)

    @Interface.private_api()
    def walker_list(self, snt: Sentinel = None, detailed: bool = False):
        """
        List walkers known to sentinel
        """
        walks = []
        for i in snt.arch_ids.obj_list():
            if i.kind == "walker":
                walks.append(i.serialize(detailed=detailed))
        return walks

    @Interface.private_api(cli_args=["name"])
    def walker_spawn_create(self, name: str, snt: Sentinel = None):
        """
        Creates new instance of walker and returns new walker object
        """
        wlk = snt.run_architype(name=name, kind="walker", caller=self, is_async=False)
        if wlk:
            if self.spawned_walker_ids.has_obj_by_name(name):
                self.spawned_walker_ids.destroy_obj_by_name(name)
            self.spawned_walker_ids.add_obj(wlk)
            self.alias_register(f"spawned:walker:{name}", wlk.jid)
            return wlk.serialize()
        else:
            return ["Walker not found!"]

    @Interface.private_api()
    def walker_spawn_list(self, detailed: bool = False):
        """
        List walkers spawned by master
        """
        walks = []
        for i in self.spawned_walker_ids.obj_list():
            walks.append(i.serialize(detailed=detailed))
        return walks

    @Interface.private_api(cli_args=["name"])
    def walker_spawn_delete(self, name: str):
        """
        Delete instance of walker
        """
        if self.spawned_walker_ids.has_obj_by_name(name):
            self.spawned_walker_ids.destroy_obj_by_name(name)
            self.alias_delete(f"spawned:walker:{name}")
            return [f"Walker {name} deteled!"]
        else:
            return [f"Walker {name} not found!"]

    @Interface.private_api()
    def walker_spawn_clear(self):
        """
        Delete instance of walker
        """
        ret = {"success": True}
        count = len(self.spawned_walker_ids)
        for i in self.spawned_walker_ids.obj_list():
            self.spawned_walker_ids.destroy_obj(i)
        ret["response"] = f"All {count} spawned walkers destroyed."
        return ret

    @Interface.private_api()
    def walker_yield_list(self, detailed: bool = False):
        """
        List walkers spawned by master
        """
        walks = []
        for i in self.yielded_walkers_ids.obj_list():
            walks.append(i.serialize(detailed=detailed))
        return walks

    @Interface.private_api(cli_args=["name"])
    def walker_yield_delete(self, name: str):
        """
        Delete instance of walker
        """
        ret = {"success": True}
        if self.yielded_walkers_ids.has_obj_by_name(name):
            self.yielded_walkers_ids.destroy_obj_by_name(name)
            ret["response"] = f"Walker {name} deteled!"
        else:
            ret["success"] = False
            ret["response"] = f"Walker {name} not found!"
        return ret

    @Interface.private_api()
    def walker_yield_clear(self):
        """
        Delete instance of walker
        """
        ret = {"success": True}
        count = len(self.yielded_walkers_ids)
        for i in self.yielded_walkers_ids.obj_list():
            self.yielded_walkers_ids.destroy_obj(i)
        ret["response"] = f"All {count} yielded walkers destroyed."
        return ret

    @Interface.private_api(cli_args=["wlk"])
    def walker_prime(
        self, wlk: Walker, nd: Node = None, ctx: dict = {}, _req_ctx: dict = {}
    ):
        """
        Assigns walker to a graph node and primes walker for execution
        """
        wlk.prime(nd, prime_ctx=ctx, request_ctx=_req_ctx)
        return [f"Walker primed on node {nd.id}"]

    @Interface.private_api(cli_args=["wlk"])
    def walker_execute(
        self,
        wlk: Walker,
        prime: Node = None,
        ctx: dict = {},
        _req_ctx: dict = {},
        profiling: bool = False,
    ):
        """
        Executes walker (assumes walker is primed)
        """
        return wlk.run(
            start_node=prime, prime_ctx=ctx, request_ctx=_req_ctx, profiling=profiling
        )

    @Interface.private_api(cli_args=["name"])
    def walker_run(
        self,
        name: str,
        nd: Node = None,
        ctx: dict = {},
        _req_ctx: dict = {},
        snt: Sentinel = None,
        profiling: bool = False,
        is_async: bool = None,
    ):
        """
        Creates walker instance, primes walker on node, executes walker,
        reports results, and cleans up walker instance.
        """
        wlk = self.yielded_walkers_ids.get_obj_by_name(name, silent=True)
        if wlk is None:
            wlk = snt.run_architype(
                name=name, kind="walker", caller=self, is_async=is_async
            )
        if wlk is None:
            return self.bad_walk_response([f"Walker {name} not found!"])
        res = self.walker_execute(
            wlk=wlk, prime=nd, ctx=ctx, _req_ctx=_req_ctx, profiling=profiling
        )
        wlk.register_yield_or_destroy(self.yielded_walkers_ids)
        return res

    @Interface.private_api(cli_args=["name"], url_args=["name"])
    def wapi(
        self,
        name: str,
        nd: Node = None,
        ctx: dict = {},
        _req_ctx: dict = {},
        snt: Sentinel = None,
        profiling: bool = False,
    ):
        """
        Walker individual APIs
        """
        return self.walker_run(name, nd, ctx, _req_ctx, snt, profiling)

    @Interface.public_api(cli_args=["wlk"])
    def walker_summon(
        self,
        key: str,
        wlk: Walker,
        nd: Node,
        ctx: dict = {},
        _req_ctx: dict = {},
        global_sync: bool = True,
    ):
        """
        Public api for running walkers, namespace key must be provided
        along with the walker id and node id
        """
        if key not in wlk.namespace_keys().values():
            return self.bad_walk_response(["Not authorized to execute this walker"])
        if global_sync:
            # Deprecated Walkers are always synced with sentinel now
            pass

        walk = wlk.duplicate()
        res = self.walker_execute(
            wlk=walk, prime=nd, ctx=ctx, _req_ctx=_req_ctx, profiling=False
        )
        walk.destroy()
        return res

    @Interface.public_api(url_args=["nd", "wlk"], allowed_methods=["post", "get"])
    def walker_callback(
        self,
        nd: Node,
        wlk: Walker,
        key: str,
        ctx: dict = {},
        _req_ctx: dict = {},
        global_sync: bool = True,
    ):
        """
        Public api for running walkers, namespace key must be provided
        along with the walker id and node id
        """

        return self.walker_summon(
            nd=nd,
            wlk=wlk,
            key=key,
            ctx=ctx,
            _req_ctx=_req_ctx,
            global_sync=global_sync,
        )

    def destroy(self):
        """
        Destroys self from memory and persistent storage
        """
        for i in self.spawned_walker_ids.obj_list():
            i.destroy()

    def bad_walk_response(self, errors=list()):
        return {"report": [], "success": False, "errors": errors}



import io
import sys
from unittest import TestCase

import jaseci.jac.tests.book_code as jtc
from jaseci.actor.sentinel import Sentinel
from jaseci.graph.graph import Graph
from jaseci.svc import MetaService
from jaseci.utils.utils import TestCaseHelper


class JacBookTests(TestCaseHelper, TestCase):
    """Unit tests for Jac language"""

    def setUp(self):
        super().setUp()
        self.sent = Sentinel(m_id="anon", h=MetaService().build_hook())
        self.gph = Graph(m_id="anon", h=self.sent._h)
        self.old_stdout = sys.stdout
        self.new_stdout = io.StringIO()
        sys.stdout = self.new_stdout

    def tearDown(self):
        sys.stdout = self.old_stdout
        super().tearDown()

    def to_screen(self):
        sys.stdout = self.old_stdout
        print("output: ", self.new_stdout.getvalue())
        sys.stdout = self.new_stdout

    def test_basic_arith(self):
        self.sent.register_code(jtc.basic_arith)
        gen_walker = self.sent.run_architype("init")
        gen_walker.prime(self.gph)
        gen_walker.run()
        self.assertEqual(self.new_stdout.getvalue(), "8 -20 1.0 -2 -13.0\n")

    def test_additional_arith(self):
        self.sent.register_code(jtc.more_arith)
        gen_walker = self.sent.run_architype("init")
        gen_walker.prime(self.gph)
        gen_walker.run()
        self.assertEqual(self.new_stdout.getvalue(), "256 4\n")

    def test_compare(self):
        self.sent.register_code(jtc.compare)
        gen_walker = self.sent.run_architype("init")
        gen_walker.prime(self.gph)
        gen_walker.run()
        self.assertEqual(
            self.new_stdout.getvalue(), "false true true false true false true\n"
        )

    def test_logical(self):
        self.sent.register_code(jtc.logical)
        gen_walker = self.sent.run_architype("init")
        gen_walker.prime(self.gph)
        gen_walker.run()
        self.assertEqual(
            self.new_stdout.getvalue(), "true false false true false true false true\n"
        )

    def test_assignments(self):
        self.sent.register_code(jtc.assignments)
        gen_walker = self.sent.run_architype("init")
        gen_walker.prime(self.gph)
        gen_walker.run()
        self.assertEqual(self.new_stdout.getvalue(), "8\n16\n36\n36.0\n-18.0\n")

    def test_if_stmt(self):
        self.sent.register_code(jtc.if_stmt)
        gen_walker = self.sent.run_architype("init")
        gen_walker.prime(self.gph)
        gen_walker.run()
        self.assertEqual(self.new_stdout.getvalue(), "Hello!\n")

    def test_else_stmt(self):
        self.sent.register_code(jtc.else_stmt)
        gen_walker = self.sent.run_architype("init")
        gen_walker.prime(self.gph)
        gen_walker.run()
        self.assertEqual(self.new_stdout.getvalue(), "A is not equal to B\n")

    def test_elif_stmt(self):
        self.sent.register_code(jtc.elif_stmt)
        gen_walker = self.sent.run_architype("init")
        gen_walker.prime(self.gph)
        gen_walker.run()
        self.assertEqual(self.new_stdout.getvalue(), "A is one less than B\n")

    def test_for_stmt(self):
        self.sent.register_code(jtc.for_stmt)
        gen_walker = self.sent.run_architype("init")
        gen_walker.prime(self.gph)
        gen_walker.run()
        self.assertEqual(
            self.new_stdout.getvalue(),
            "Hello 0 times!\nHello 1 times!\nHello 2 times!\n"
            "Hello 3 times!\nHello 4 times!\nHello 5 times!\n"
            "Hello 6 times!\nHello 7 times!\nHello 8 times!\n"
            "Hello 9 times!\n",
        )

    def test_while_stmt(self):
        self.sent.register_code(jtc.while_stmt)
        gen_walker = self.sent.run_architype("init")
        gen_walker.prime(self.gph)
        gen_walker.run()
        self.assertEqual(
            self.new_stdout.getvalue(),
            "Hello 5 times!\nHello 4 times!\nHello 3 times!\n"
            "Hello 2 times!\nHello 1 times!\n",
        )

    def test_break_stmt(self):
        self.sent.register_code(jtc.break_stmt)
        gen_walker = self.sent.run_architype("init")
        gen_walker.prime(self.gph)
        gen_walker.run()
        self.assertEqual(
            self.new_stdout.getvalue(),
            "Hello 0 times!\nHello 1 times!\nHello 2 times!\n"
            "Hello 3 times!\nHello 4 times!\nHello 5 times!\n"
            "Hello 6 times!\n",
        )

    def test_continue_stmt(self):
        self.sent.register_code(jtc.continue_stmt)
        gen_walker = self.sent.run_architype("init")
        gen_walker.prime(self.gph)
        gen_walker.run()
        self.assertEqual(
            self.new_stdout.getvalue(),
            "Hello 5 times!\nHello 4 times!\n" "Hello 2 times!\nHello 1 times!\n",
        )

    def test_continue_stmt2(self):
        self.sent.register_code(jtc.continue_stmt2)
        gen_walker = self.sent.run_architype("init")
        gen_walker.prime(self.gph)
        gen_walker.run()
        self.assertEqual(
            self.new_stdout.getvalue(),
            "hello 9\nhello 8\n" "hello 7\nhello 6\nhello 5\n",
        )

    def test_destroy_disconnect(self):
        self.sent.register_code(jtc.destroy_disconn)
        gen_walker = self.sent.run_architype("init")
        gen_walker.prime(self.gph)
        gen_walker.run()
        out = self.new_stdout.getvalue()
        self.assertEqual(out.count('"'), 10)

    def test_array_assign(self):
        self.sent.register_code(jtc.array_assign)
        gen_walker = self.sent.run_architype("init")
        gen_walker.prime(self.gph)
        gen_walker.run()
        self.assertEqual(
            self.new_stdout.getvalue(),
            "[[0, 0], [0, 0]]\n"
            "[[1, 1], [0, 0]]\n"
            "[[1, 2], [3, 4]]\n"
            "[[1, 2], [1, 2]]\n",
        )

    def test_md_array_assign(self):
        self.sent.register_code(jtc.array_md_assign)
        gen_walker = self.sent.run_architype("init")
        gen_walker.prime(self.gph)
        gen_walker.run()
        self.assertEqual(
            self.new_stdout.getvalue(), "[[1, 2], [3, 4]]\n" "[[1, 1], [3, 4]]\n"
        )

    def test_dereference(self):
        self.sent.register_code(jtc.dereference)
        gen_walker = self.sent.run_architype("init")
        gen_walker.prime(self.gph)
        gen_walker.run()
        self.assertEqual(self.new_stdout.getvalue()[:9], "urn:uuid:")

    def test_pre_post_walk(self):
        self.sent.register_code(jtc.pre_post_walking)
        gen_walker = self.sent.run_architype("init")
        gen_walker.prime(self.gph)
        gen_walker.run()
        self.assertEqual(self.new_stdout.getvalue(), "count: 8\n")

    def test_pre_post_walk_disengage(self):
        self.sent.register_code(jtc.pre_post_walking_dis)
        gen_walker = self.sent.run_architype("init")
        gen_walker.prime(self.gph)
        gen_walker.run()
        self.assertEqual(self.new_stdout.getvalue(), "count: 6\n")

    def test_length(self):
        self.sent.register_code(jtc.length)
        gen_walker = self.sent.run_architype("init")
        gen_walker.prime(self.gph)
        gen_walker.run()
        self.assertEqual(self.new_stdout.getvalue(), "3\n3\n")

    def test_sort_by_col(self):
        self.sent.register_code(jtc.sort_by_col)
        gen_walker = self.sent.run_architype("init")
        gen_walker.prime(self.gph)
        gen_walker.run()
        self.assertEqual(
            self.new_stdout.getvalue(),
            '[["b", 333], ["c", 245], ["a", 56]]\n'
            '[["a", 56], ["b", 333], ["c", 245]]\n'
            '[["c", 245], ["b", 333], ["a", 56]]\n'
            '[["a", 56], ["c", 245], ["b", 333]]\n'
            '[["b", 333], ["c", 245], ["a", 56]]\n',
        )

    def test_list_remove_element(self):
        self.sent.register_code(jtc.list_remove)
        gen_walker = self.sent.run_architype("init")
        gen_walker.prime(self.gph)
        gen_walker.run()
        self.assertEqual(
            self.new_stdout.getvalue(),
            '[["b", 333], ["c", 245], ["a", 56]]\n'
            '[["b", 333], ["a", 56]]\n[["b", 333]]\n',
        )

    def test_can_action(self):
        self.sent.register_code(jtc.can_action)
        gen_walker = self.sent.run_architype("init")
        gen_walker.prime(self.gph)
        gen_walker.run()
        self.assertEqual(self.new_stdout.getvalue(), "56 7\n" "56 8\n")

    def test_can_action_param(self):
        self.sent.register_code(jtc.can_action_params)
        gen_walker = self.sent.run_architype("init")
        gen_walker.prime(self.gph)
        gen_walker.run()
        self.assertEqual(
            self.new_stdout.getvalue(), "43 7\n" "43 8\n" "48 7\n" "48 8\n"
        )

    def test_cross_scope_report(self):
        self.sent.register_code(jtc.cross_scope_report)
        gen_walker = self.sent.run_architype("init")
        gen_walker.prime(self.gph)
        gen_walker.run()
        self.assertIn(56, gen_walker.report)
        self.assertIn(8, gen_walker.report)
        self.assertEqual(self.new_stdout.getvalue(), "56 7\n" "56 8\n")

    def test_has_private(self):
        self.sent.register_code(jtc.has_private)
        gen_walker = self.sent.run_architype("init")
        gen_walker.prime(self.gph)
        gen_walker.run()
        print(gen_walker.report)
        self.assertIn("'context': {'apple': 12, 'grape': 1", self.new_stdout.getvalue())

    def test_array_idx_of_expr(self):
        self.sent.register_code(jtc.array_idx_of_expr)
        gen_walker = self.sent.run_architype("init")
        gen_walker.prime(self.gph)
        gen_walker.run()
        self.assertEqual(self.new_stdout.getvalue(), "3\n1\n")

    def test_dict_assign(self):
        self.sent.register_code(jtc.dict_assign)
        gen_walker = self.sent.run_architype("init")
        gen_walker.prime(self.gph)
        gen_walker.run()
        self.assertEqual(
            self.new_stdout.getvalue(),
            '{"three": 3, "four": 4}\n'
            '{"three": 3, "four": 55}\n'
            '{"one": 1, "two": 2}\n2\n',
        )

    def test_dict_md_assign(self):
        self.sent.register_code(jtc.dict_md_assign)
        gen_walker = self.sent.run_architype("init")
        gen_walker.prime(self.gph)
        gen_walker.run()
        self.assertEqual(
            self.new_stdout.getvalue(),
            '{"one": {"inner": 44}, "two": 2}\n' '{"inner": 2}\n2\n',
        )

    def test_dict_keys(self):
        self.sent.register_code(jtc.dict_keys)
        gen_walker = self.sent.run_architype("init")
        gen_walker.prime(self.gph)
        gen_walker.run()
        self.assertEqual(
            self.new_stdout.getvalue(),
            '{"one": {"inner": 44}, "two": 2}\n' '{"inner": 2}\n2\n',
        )

    def test_cond_dict_keys(self):
        self.sent.register_code(jtc.cond_dict_keys)
        gen_walker = self.sent.run_architype("init")
        gen_walker.prime(self.gph)
        gen_walker.run()
        self.assertEqual(
            self.new_stdout.getvalue(),
            '{"one": {"inner": 44}, "two": 2}\n' "is here\n" "also not here\n",
        )

    def test_vector_softmax(self):
        self.sent.register_code(jtc.soft_max)
        gen_walker = self.sent.run_architype("init")
        gen_walker.prime(self.gph)
        gen_walker.run()
        self.assertEqual(
            self.new_stdout.getvalue(),
            "[0.8360188027814407, 0.11314284146556013, 0.05083835575299916]\n",
        )

    def test_book_fam_example(self):
        self.sent.register_code(jtc.fam_example)
        gen_walker = self.sent.run_architype("create_fam")
        gen_walker.prime(self.gph)
        gen_walker.run()
        self.assertEqual(
            self.new_stdout.getvalue(), "I didn't do any of the hard work.\n"
        )

    def test_book_visitor_preset(self):
        self.sent.register_code(jtc.visitor_preset)
        gen_walker = self.sent.run_architype("init")
        gen_walker.prime(self.gph)
        gen_walker.run()
        outsplit = self.new_stdout.getvalue().split("\n")
        self.assertIn("from", outsplit[0])
        self.assertIn("setter", outsplit[0])
        self.assertIn("walker", outsplit[0])
        self.assertIn("init", outsplit[1])
        self.assertIn("walker", outsplit[1])
        self.assertIn("init only", outsplit[2])
        self.assertIn('"name": "init"', outsplit[2])

    def test_book_visitor_local_aciton(self):
        self.sent.register_code(jtc.visitor_local_aciton)
        gen_walker = self.sent.run_architype("init")
        gen_walker.prime(self.gph)
        gen_walker.run()
        outsplit = self.new_stdout.getvalue().split("\n")
        self.assertIn('"byear": null', outsplit[0])
        self.assertIn("to 1995: {", outsplit[1])
        self.assertIn("setter", outsplit[2])
        self.assertIn('"byear": "1995-01-01"', outsplit[3])
        self.assertIn('"name": "init"', outsplit[4])

    def test_book_copy_assign_to_edge(self):
        self.sent.register_code(jtc.copy_assign_to_edge)
        gen_walker = self.sent.run_architype("init")
        gen_walker.prime(self.gph)
        gen_walker.run()
        outsplit = self.new_stdout.getvalue().split("\n")
        self.assertIn('"Josh", "age": 32', outsplit[1])
        self.assertIn("college", outsplit[2])
        self.assertIn('"Jane", "age": 30', outsplit[3])
        self.assertIn("sister", outsplit[4])
        self.assertIn('"Josh", "age": 32', outsplit[5])
        self.assertIn("college", outsplit[6])
        self.assertIn('"Jane", "age": 30', outsplit[7])
        self.assertIn("sister", outsplit[8])



import ssl
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from json import dumps
from smtplib import SMTP, SMTP_SSL

from jaseci.svc import CommonService, ServiceState as Ss
from jaseci.utils.utils import logger
from .common import MAIL_CONFIG


#################################################
#                  EMAIL APP                   #
#################################################


class MailService(CommonService):

    ###################################################
    #                   INITIALIZER                   #
    ###################################################

    def __init__(self, hook=None):
        super().__init__(MailService)

        try:
            if self.is_ready():
                self.state = Ss.STARTED
                self.__mail(hook)
        except Exception as e:
            if not (self.quiet):
                logger.error(
                    "Skipping Mail setup due to initialization failure!\n"
                    f"{e.__class__.__name__}: {e}"
                )
            self.app = None
            self.state = Ss.FAILED

    def __mail(self, hook):
        configs = self.get_config(hook)
        enabled = configs.get("enabled", True)

        if enabled:
            self.quiet = configs.pop("quiet", False)
            self.__convert_config(hook, configs)
            self.app = self.connect(configs)
            self.state = Ss.RUNNING
        else:
            self.state = Ss.DISABLED

    # ----------- BACKWARD COMPATIBILITY ------------ #
    # ---------------- TO BE REMOVED ---------------- #

    def __convert(self, hook, holder, mapping: dict):
        for k, v in mapping.items():
            if hook.has_glob(k):
                conf = hook.get_glob(k)
                if not (conf is None):
                    if v == "tls":
                        holder[v] = conf.lower() == "true"
                    else:
                        holder[v] = conf
        return holder

    def __convert_config(self, hook, configs: dict):
        version = configs.get("version", 2)
        migrate = configs.pop("migrate", False)
        if version == 1 or migrate:
            self.__convert(
                hook,
                configs,
                {
                    "EMAIL_BACKEND": "backend",
                    "EMAIL_HOST": "host",
                    "EMAIL_PORT": "port",
                    "EMAIL_HOST_USER": "user",
                    "EMAIL_HOST_PASSWORD": "pass",
                    "EMAIL_DEFAULT_FROM": "sender",
                    "EMAIL_USE_TLS": "tls",
                },
            )

            if "templates" not in configs:
                configs["templates"] = {}

            self.__convert(
                hook,
                configs["templates"],
                {
                    "EMAIL_ACTIVATION_SUBJ": "activation_subj",
                    "EMAIL_ACTIVATION_BODY": "activation_body",
                    "EMAIL_ACTIVATION_HTML_BODY": "activation_html_body",
                    "EMAIL_RESETPASS_SUBJ": "resetpass_subj",
                    "EMAIL_RESETPASS_BODY": "resetpass_body",
                    "EMAIL_RESETPASS_HTML_BODY": "resetpass_html_body",
                },
            )

            if migrate:
                hook.save_glob("MAIL_CONFIG", dumps(configs))

    ###################################################
    #                     CLEANER                     #
    ###################################################

    def reset(self, hook):
        if self.is_running():
            self.app.terminate()
        self.build(hook)

    ####################################################
    #                    OVERRIDDEN                    #
    ####################################################

    def connect(self, configs):
        host = configs.get("host")
        port = configs.get("port")
        user = configs.get("user")
        _pass = configs.get("pass")
        sender = configs.get("sender", user)

        context = ssl.create_default_context()

        if configs.get("tls", True):
            server = SMTP(host, port)
            server.ehlo()
            server.starttls(context=context)
            server.ehlo()
        else:
            server = SMTP_SSL(host, port, context=context)

        server.login(user, _pass)

        return Mailer(server, sender)

    def get_config(self, hook) -> dict:
        return hook.build_config("MAIL_CONFIG", MAIL_CONFIG)


# ----------------------------------------------- #


####################################################
#                   EMAIL CONFIG                   #
####################################################


class Mailer:
    def __init__(self, server, sender):
        self.server = server
        self.sender = sender

    def send_custom_email(
        self,
        sender: str = None,
        recipients: list = [],
        subject: str = "Jaseci Email",
        body: tuple = ("", ""),
    ):
        message = MIMEMultipart()
        message["Subject"] = subject
        message["From"] = self.sender if sender is None else sender
        message["To"] = ", ".join(recipients)

        message.attach(MIMEText(body[0], "plain"))
        message.attach(MIMEText(body[1], "html"))
        self.server.sendmail(message["From"], recipients, message.as_string())

    def terminate(self):
        self.server.quit()



from prometheus_api_client import PrometheusConnect


class Promon:
    def __init__(self, url: str):
        self.prom = PrometheusConnect(url=url, disable_ssl=True)

    def all_metrics(self) -> list:
        return self.prom.all_metrics()

    def cpu_utilization_core(self) -> dict:
        util = self.prom.get_current_metric_value(
            'sum(irate(node_cpu_seconds_total{mode!="idle"}[10m])) by (node)'
        )
        res = {}
        for node in util:
            node_name = node["metric"]["node"]
            node_util = float(node["value"][1])
            res[node_name] = node_util
        return res

    def cpu_utilization_percentage(self) -> dict:
        util = self.prom.get_current_metric_value(
            '(sum(irate(node_cpu_seconds_total{mode!="idle"}[10m])) by '
            '(node)) / (sum(irate(node_cpu_seconds_total{mode!=""}[10m])) by '
            "(node)) * 100"
        )
        res = {}
        for node in util:
            node_name = node["metric"]["node"]
            node_util = float(node["value"][1])
            res[node_name] = node_util
        return res

    def cpu_utilization_per_pod_cores(self) -> dict:
        util = self.prom.get_current_metric_value(
            'sum(irate(container_cpu_usage_seconds_total{pod!=""}[10m])) by (pod)'
        )
        res = {}
        for pod in util:
            pod_name = pod["metric"]["pod"]
            value = float(pod["value"][1])
            res[pod_name] = float(value)
        return res

    def mem_total_bytes(self) -> dict:
        util = self.prom.get_current_metric_value(
            "sum(node_memory_MemTotal_bytes) by (node)"
        )
        res = {}
        for node in util:
            node_name = node["metric"]["node"]
            node_util = float(node["value"][1])
            res[node_name] = node_util
        return res

    def mem_utilization_bytes(self) -> dict:
        util = self.prom.get_current_metric_value(
            "sum(node_memory_Active_bytes) by (node)"
        )
        res = {}
        for node in util:
            node_name = node["metric"]["node"]
            node_util = float(node["value"][1])
            res[node_name] = node_util
        return res

    def mem_utilization_percentage(self) -> dict:
        util = self.prom.get_current_metric_value(
            "sum(node_memory_Active_bytes / node_memory_MemTotal_bytes * 100 ) by "
            "(node)"
        )
        res = {}
        for node in util:
            node_name = node["metric"]["node"]
            node_util = float(node["value"][1])
            res[node_name] = node_util
        return res

    def mem_utilization_per_pod_bytes(self) -> dict:
        util = self.prom.get_current_metric_value(
            'sum(container_memory_working_set_bytes{pod!=""}) by (pod)'
        )
        res = {}
        for pod in util:
            pod_name = pod["metric"]["pod"]
            value = float(pod["value"][1])
            res[pod_name] = float(value)
        return res

    def node_pods(self) -> dict:
        util = self.prom.get_current_metric_value("kube_pod_info")
        res = {}
        for pod in util:
            info = pod["metric"]
            node = info["node"]
            pod = info["pod"]
            if res.get(node) is None:
                res[node] = set()
            res[node].add(pod)
        return res

    def network_receive_bytes(self) -> dict:
        util = self.prom.get_current_metric_value(
            "sum (rate (node_network_receive_bytes_total{}[10m])) by (node)"
        )
        res = {}
        for node in util:
            node_name = node["metric"]["node"]
            node_util = float(node["value"][1])
            res[node_name] = node_util
        return res

    def network_receive_per_pod_bytes(self):
        util = self.prom.get_current_metric_value(
            'sum (rate (container_network_receive_bytes_total{pod!=""}[10m])) by (pod)'
        )
        res = {}
        for pod in util:
            pod_name = pod["metric"]["pod"]
            value = pod["value"][1]
            res[pod_name] = float(value)
        return res

    def network_transmit_bytes(self) -> dict:
        util = self.prom.get_current_metric_value(
            "sum (rate (node_network_transmit_bytes_total{}[10m])) by (node)"
        )
        res = {}
        for node in util:
            node_name = node["metric"]["node"]
            node_util = float(node["value"][1])
            res[node_name] = node_util
        return res

    def network_transmit_per_pod_bytes(self):
        util = self.prom.get_current_metric_value(
            'sum (rate (container_network_transmit_bytes_total{pod!=""}[10m])) by (pod)'
        )
        res = {}
        for pod in util:
            pod_name = pod["metric"]["pod"]
            value = pod["value"][1]
            res[pod_name] = float(value)
        return res

    def disk_total_bytes(self) -> dict:
        util = self.prom.get_current_metric_value(
            'sum(avg (node_filesystem_size_bytes{mountpoint!="/boot", '
            'fstype!="tmpfs"}) without (mountpoint)) by (node)'
        )
        res = {}
        for node in util:
            node_name = node["metric"]["node"]
            node_util = float(node["value"][1])
            res[node_name] = node_util
        return res

    def disk_free_bytes(self) -> dict:
        util = self.prom.get_current_metric_value(
            'sum(avg (node_filesystem_free_bytes{mountpoint!="/boot", '
            'fstype!="tmpfs"}) without (mountpoint)) by (node)'
        )
        res = {}
        for node in util:
            node_name = node["metric"]["node"]
            node_util = float(node["value"][1])
            res[node_name] = node_util
        return res

    def pod_info(self) -> dict:
        util = self.prom.get_current_metric_value("kube_pod_info")
        res = {}
        for pod in util:
            pod_name = pod["metric"]["pod"]
            res[pod_name] = pod["metric"]

        cpu = self.cpu_utilization_per_pod_cores()
        for pod in util:
            pod_name = pod["metric"]["pod"]
            pod_cpu = cpu.get(pod_name, 0)
            res[pod_name]["cpu_utilization_cores"] = pod_cpu

        mem = self.mem_utilization_per_pod_bytes()
        for pod in util:
            pod_name = pod["metric"]["pod"]
            pod_mem = mem.get(pod_name, 0)
            res[pod_name]["mem_utilization_bytes"] = pod_mem

        recv = self.network_receive_per_pod_bytes()
        for pod in util:
            pod_name = pod["metric"]["pod"]
            pod_recv = recv.get(pod_name, 0)
            res[pod_name]["network_recv_bytes"] = pod_recv

        tran = self.network_transmit_per_pod_bytes()
        for pod in util:
            pod_name = pod["metric"]["pod"]
            pod_tran = tran.get(pod_name, 0)
            res[pod_name]["network_tran_bytes"] = pod_tran

        return res



from redis import Redis

from jaseci.svc import CommonService, ServiceState as Ss
from jaseci.utils.utils import logger
from .common import REDIS_CONFIG


#################################################
#                  REDIS HOOK                   #
#################################################


class RedisService(CommonService):

    ###################################################
    #                   INITIALIZER                   #
    ###################################################

    def __init__(self, hook=None):
        super().__init__(RedisService)

        try:
            if self.is_ready():
                self.state = Ss.STARTED
                self.__redis(hook)
        except Exception as e:
            if not (self.quiet):
                logger.error(
                    "Skipping Redis due to initialization failure!\n"
                    f"{e.__class__.__name__}: {e}"
                )
            self.app = None
            self.state = Ss.FAILED

    def __redis(self, hook):
        configs = self.get_config(hook)
        enabled = configs.pop("enabled", True)

        if enabled:
            self.quiet = configs.pop("quiet", False)
            self.app = Redis(**configs, decode_responses=True)
            self.app.ping()
            self.state = Ss.RUNNING
        else:
            self.state = Ss.DISABLED

    ###################################################
    #                     COMMONS                     #
    ###################################################

    def get(self, name):
        return self.app.get(name)

    def set(self, name, val):
        self.app.set(name, val)

    def exists(self, name):
        return self.app.exists(name)

    def delete(self, name):
        self.app.delete(name)

    def hget(self, name, key):
        return self.app.hget(name, key)

    def hset(self, name, key, val):
        self.app.hset(name, key, val)

    def hexists(self, name, key):
        return self.app.hexists(name, key)

    def hdel(self, name, key):
        self.app.hdel(name, key)

    def hkeys(self, name):
        return self.app.hkeys(name)

    ###################################################
    #                     CLEANER                     #
    ###################################################

    def reset(self, hook):
        self.build(hook)

    def clear(self):
        if self.is_running():
            self.app.flushdb()

    ###################################################
    #                     CONFIG                      #
    ###################################################

    def get_config(self, hook) -> dict:
        return hook.build_config("REDIS_CONFIG", REDIS_CONFIG)


# ----------------------------------------------- #



"""
Mix in for jac code object in Jaseci
"""
import json
from jaseci.utils.utils import logger
from jaseci.jac.ir.ast_builder import JacAstBuilder
from jaseci.jac.ir.ast import Ast
import hashlib
from pathlib import Path
from os.path import dirname

# Used to check ir matches grammar of current Jaseci instance
grammar_hash = hashlib.md5(
    Path(dirname(__file__) + "/../jac.g4").read_text().encode()
).hexdigest()


class JacJsonEnc(json.JSONEncoder):
    """Custom Json encoder for Jac ASTs"""

    def default(self, obj):
        if isinstance(obj, Ast):
            retd = {}
            for i in obj.__dict__.keys():
                if not i.startswith("_"):
                    retd[i] = obj.__dict__[i]
            return retd
        return super().default(obj)


class JacJsonDec(json.JSONDecoder):
    """Custom hook for decoding Jac ASTs"""

    def __init__(self, *args, **kwargs):
        json.JSONDecoder.__init__(self, object_hook=self.object_hook, *args, **kwargs)

    def object_hook(self, obj):

        if isinstance(obj, dict) and "loc" in obj and "kid" in obj:
            ret = Ast(mod_name=obj["loc"][2])
            for i in obj.keys():
                setattr(ret, i, obj[i])
            return ret
        return obj


def jac_ast_to_ir(jac_ast: Ast):
    """Convert AST to IR string"""
    return json.dumps(cls=JacJsonEnc, obj={"gram_hash": grammar_hash, "ir": jac_ast})


def jac_ir_to_ast(ir: str):
    """Convert IR string to AST"""
    ir_load = json.loads(cls=JacJsonDec, s=ir)
    if (
        not isinstance(ir_load, dict)
        or "gram_hash" not in ir_load
        or ir_load["gram_hash"] != grammar_hash
        or (not isinstance(ir_load["ir"], Ast) and not ir_load["ir"] is None)
    ):
        logger.error(
            "Jac IR invalid or incompatible with current Jaseci "
            f"(valid gram_hash: {grammar_hash}!"
        )
    return ir_load["ir"]


class JacCode:
    """Obj mixin to code pickling"""

    def __init__(self, code_ir=None):
        self.is_active = False
        self.code_ir = None
        self.code_sig = None
        self._jac_ast = None
        self.errors = []
        self.apply_ir(code_ir)

    def reset(self):
        JacCode.__init__(self)

    def refresh(self):
        self._jac_ast = jac_ir_to_ast(self.code_ir) if self.code_ir else None
        if self._jac_ast:
            self.is_active = True
        else:
            self.is_active = False

    def apply_ir(self, ir):
        """Apply's IR to object"""
        self.code_ir = (
            ir.strip()
            if (isinstance(ir, str))
            else json.dumps(ir)
            if (isinstance(ir, dict))
            else jac_ast_to_ir(ir)
        )
        self.code_sig = hashlib.md5(self.code_ir.encode()).hexdigest()
        JacCode.refresh(self)  # should disregard overloaded versions

    def parse_jac(self, code, dir, start_rule="start"):
        """Generate AST tree from Jac code text"""
        tree = JacAstBuilder(
            jac_text=code, start_rule=start_rule, mod_name=self.name, mod_dir=dir
        )
        self.errors = tree._parse_errors
        if tree._parse_errors:
            logger.error(str(f"{self.name}: Invalid syntax in Jac code!"))
            for i in tree._parse_errors:
                logger.error(i)
            return None
        return tree.root

    def register(self, code, dir):
        """
        Parses Jac code and saves IR
        """
        start_rule = "start" if self.j_type == "sentinel" else self.j_type
        tree = self.parse_jac(code, dir, start_rule=start_rule)

        if not tree:
            self.is_active = False
        else:
            self.apply_ir(tree)

        if not self.is_active:
            logger.error(str(f"{self.name}: Code not registered"))
        return self.is_active

    def ir_dict(self):
        """Return IR as dictionary"""
        return json.loads(self.code_ir)



"""
Super (master) api as a mixin
"""
from jaseci.api.interface import Interface
from jaseci.element.master import Master


class SuperApi:
    """Super APIs for creating nicknames for UUIDs and other long strings"""

    @Interface.admin_api(cli_args=["name"])
    def master_createsuper(
        self,
        name: str,
        global_init: str = "",
        global_init_ctx: dict = {},
        other_fields: dict = {},
    ):
        """
        Create a super instance and return root node super object

        other_fields used for additional feilds for overloaded interfaces
        (i.e., Dango interface)
        """
        if self.sub_master_ids.has_obj_by_name(name):
            return {"response": f"{name} already exists", "success": False}
        ret = {}
        mast = self.superuser_creator(name, other_fields)
        ret["user"] = mast.serialize()
        if len(global_init):
            ret["global_init"] = self.user_global_init(
                mast, global_init, global_init_ctx
            )
        self.take_ownership(mast)
        ret["success"] = True
        return ret

    @Interface.admin_api()
    def master_allusers(self, limit: int = 0, offset: int = 0, asc: bool = False):
        """
        Returns info on a set of users, limit specifies the number of users to
        return and offset specfies where to start
        NOTE: Abstract interface to be overridden
        """

    @Interface.admin_api(cli_args=["mast"])
    def master_become(self, mast: Master):
        """
        Sets the default master master should use
        """
        self.caller = mast.jid
        return {"response": f"You are now {mast.name}"}

    @Interface.admin_api(cli_args=["mast"])
    def master_unbecome(self):
        """
        Unsets the default master master should use
        """
        self.caller = None
        return {"response": f"You are now {self.name}"}



from jaseci.actions.live_actions import jaseci_action
from jaseci.svc import MetaService
from jaseci.svc.mail import MAIL_ERR_MSG
from jaseci.utils.utils import logger


@jaseci_action()
def send(sender, recipients, subject, text, html):
    ma = MetaService().get_service("mail")
    if ma.is_running():
        ma.app.send_custom_email(sender, recipients, subject, (text, html))
    else:
        logger.error(MAIL_ERR_MSG)



from .state import ServiceState
from .common import CommonService, ProxyService, MetaProperties
from .redis import RedisService
from .task import TaskService
from .mail import MailService
from .meta import MetaService

__all__ = [
    "ServiceState",
    "MetaProperties",
    "CommonService",
    "ProxyService",
    "RedisService",
    "TaskService",
    "MailService",
    "MetaService",
]



"""Built in actions for Jaseci"""
from jaseci.actions.live_actions import jaseci_action
from jaseci.utils.utils import master_from_meta, copy_func
from jaseci.element.super_master import SuperMaster
import functools


def interface_api(api_name, is_public, p_spec, *args, meta):
    """
    Interfaces Master apis after processing arguments/parameters
    from cli
    """
    mast = master_from_meta(meta)
    params = construct_params(p_spec, args)
    if is_public:
        return mast.public_interface_to_api(params, api_name)
    else:
        return mast.general_interface_to_api(params, api_name)


def construct_params(p_spec, args):
    arg_count = 0
    params = {}
    for i in p_spec.keys():
        if i == "self":
            continue
        p_default = (
            p_spec[i].default if p_spec[i].default is not p_spec[i].empty else None
        )
        if arg_count < len(args):
            params[i] = args[arg_count]
        else:
            params[i] = p_default
        arg_count += 1
    return params


def build_cmds():
    """
    Generates Click function with options for each command
    group and leaf signatures
    leaf is format: [api_name, func_sig, is_public, func_doc]
    """

    for i in SuperMaster.all_apis(None):
        func_name = "_".join(i["groups"])
        f = functools.partial(
            copy_func(interface_api, func_name),
            func_name,
            i in SuperMaster._public_api,
            i["sig"].parameters,
        )
        f.__name__ = func_name
        f.__doc__ = i["doc"]
        f.__module__ = __name__

        # original_args = inspect.getargspec(func).args
        globals()[func_name] = jaseci_action()(f)


build_cmds()



################################################
#                   DEFAULTS                   #
################################################

MAIL_ERR_MSG = "Mail service is disabled or not yet configured!"
MAIL_CONFIG = {
    "enabled": True,
    "quiet": True,
    "version": 2,
    "tls": True,
    "host": "smtp.gmail.com",
    "port": 587,
    "sender": "Jaseci Admin<boyong@jaseci.org>",
    "user": "jaseci.dev@gmail.com",
    "pass": "yrtviyrdzmzdpjxg",
    "backend": "smtp",
    "templates": {
        "activation_subj": "Please activate your account!",
        "activation_body": "Thank you for creating an account!\n\n"
        "Activation Code: {{code}}\n"
        "Please click below to activate:\n{{link}}",
        "activation_html_body": "Thank you for creating an account!<br><br>"
        "Activation Code: {{code}}<br>"
        "Please click below to activate:<br>"
        "{{link}}",
        "resetpass_subj": "Password Reset for Jaseci Account",
        "resetpass_body": "Your Jaseci password reset token is: {{token}}",
        "resetpass_html_body": "Your Jaseci password reset" "token is: {{token}}",
    },
}



from jaseci.utils.utils import obj_class_cache, build_class_dict
from jaseci.element.super_master import SuperMaster as Sm
from docstring_parser import parse
from os.path import exists

# from pprint import pformat
from inspect import getdoc, signature
import jaseci


class Book:
    def format_params(self, sig, ignore_args=[]):
        ret = ""
        for i in sig.parameters:
            if i == "self" or i in ignore_args:
                continue
            if len(ret):
                ret += ", "
            ret += i + ": " + sig.parameters[i].annotation.__name__
            default = sig.parameters[i].default
            if default == sig.parameters[i].empty:
                ret += " (*req)"
            else:
                # if isinstance(default, str):
                #     default = default.encode("unicode_escape").lstrip("b")
                default = (
                    str(default)
                    .replace("_", "\\_")
                    .replace("{", "\\{")
                    .replace("}", "\\}")
                    .replace("\n", "\\\\n")
                )
                ret += f" ({default})"
        return ret if len(ret) else "n/a"

    def bookgen_api_cheatsheet(self, root, out=None, str=""):
        if out is None:
            out = []
        if "leaf" in root.keys():
            line = "\\lstinline$" + str.strip()
            if root["leaf"][5]:  # cli_only
                line += " (cli only)"
            line += (
                "$ "
                + "& \\lstinline$"
                + f'{self.format_params(root["leaf"][1])}'
                + "$ \\\\ \\hline\n"
            )
            out.append(line)
            return
        for i in root.keys():
            self.bookgen_api_cheatsheet(root[i], out, str + f"{i} ")
        return "".join(out)

    def get_stdlib_pre_table(self):
        clip = (
            "\\rowcolors{1}{light-cyan}{light-gray}\\begin{longtable}"
            "{|p{4cm} | p{6cm}|}\\toprule\\rowcolor{white}\\textbf"
            "{Action}&\\textbf{Args}\\\\\\midrule"
        )
        return clip

    def get_stdlib_post_table(self, act="default"):
        clip = (
            "\\bottomrule\\hiderowcolors\\caption{"
            + act
            + " Actions in Jac}\\label{tab:"
            + act
            + "std}\\end{longtable}"
        )
        return clip

    def get_global_actions(self):
        import jaseci.actions.standard as stdact
        import pkgutil
        from importlib.machinery import SourceFileLoader
        from jaseci.actions.live_actions import live_actions

        all_action_sets = []
        for importer, modname, ispkg in pkgutil.iter_modules(stdact.__path__):
            if not exists(stdact.__path__[0] + "/" + modname + ".py"):
                continue
            mod = SourceFileLoader(
                modname, stdact.__path__[0] + "/" + modname + ".py"
            ).load_module()
            all_action_sets.append(
                [modname]
                + [getdoc(mod)]
                + [
                    [name, val]
                    for name, val in mod.__dict__.items()
                    if callable(val) and modname + "." + name in live_actions
                ]
            )
        return all_action_sets

    def func_to_sexy_box(self, fname, func, ignore_args=[]):
        doc = getdoc(func)
        line = (
            "\\apispec{"
            + fname
            + "}{"
            + f"{self.format_params(signature(func), ignore_args=ignore_args)}"
            + "}\n"
        )
        parsed_doc = parse(doc)
        doc = parsed_doc.long_description
        if doc is None:
            doc = "No documentation yet."
        doc = doc.replace("_", " ")
        if len(parsed_doc.params):
            # doc += "\\vspace{3mm}\\par\n\\textbf{Parameters}\n\\par"
            doc += "\\vspace{4mm}\\par\n"
            args_doc = "\\argspec{Parameters}{"
            for i in parsed_doc.params:
                args_doc += (
                    f"\n\\texttt{{{i.arg_name}}} -"
                    f"- {i.description}\\vspace{{1.5mm}}\\par\n"
                )
            args_doc += "}"
            args_doc = args_doc.replace("_", "\\_")
            doc += args_doc
        if parsed_doc.returns:
            # doc += "\\vspace{3mm}\\par\n\\textbf{Parameters}\n\\par"
            doc += "\\vspace{4mm}\\par\n"
            args_doc = "\\argspec{Returns}{" + parsed_doc.returns.description + "}"
            args_doc = args_doc.replace("_", "\\_")
            doc += args_doc
        line += "{" + doc + "}\n"
        return line

    def bookgen_std_library(self):
        out = []
        for i in self.get_global_actions():
            lib = i[0]
            moddoc = parse(i[1]).long_description
            if moddoc is None:
                moddoc = "No documentation yet."
            if lib == "jaseci":
                continue
            out += ["\\subsection{" + lib + "}\n\\par\n" + moddoc + "\n"]
            i = i[2:]
            for j in i:
                out.append(
                    self.func_to_sexy_box(
                        ".".join([lib, j[0].replace("_", "\\_")]),
                        j[1],
                        ignore_args=["meta"],
                    )
                )
        return "".join(out)

    def std_library_table(self):
        out = []
        for i in self.get_global_actions():
            lib = i[0]
            if lib == "jaseci":
                continue
            out += ["\\subsection{" + lib + "}\n", self.get_stdlib_pre_table()]

            i = i[1:]
            for j in i:
                line = (
                    "\\lstinline$"
                    + lib
                    + "."
                    + j[0]
                    + "$ & \\lstinline$"
                    + f"{self.format_params(signature(j[1]))}"
                    + "$ \\\\ \\hline\n"
                )
                out.append(line)
                doc = getdoc(j[1]) if getdoc(j[1]) is not None else ""
                line = "\\multicolumn{2}{|p|}{Description: " + doc + "} \\\\ \\hline\n"
                out.append(line)
            out.append(self.get_stdlib_post_table(lib))
        return "".join(out)

    def bookgen_api_spec(self):
        ret = ""
        build_class_dict(jaseci)
        for i in obj_class_cache.keys():
            if not i.endswith("_api"):
                continue
            ret += f"\\subsection{{APIs for {i[:-4]}}}\n\n"
            doc = getdoc(obj_class_cache[i]).replace("\n\n", "\n\\par\n")
            doc = parse(doc).long_description
            doc = (
                doc.replace("_", "\\_") if doc is not None else "No documentation yet."
            )
            ret += f"{doc}\n\n"
            ret += self.api_call_spec(obj_class_cache[i])
        return ret

    def api_call_spec(self, cls):
        ret = ""
        for i, v in cls.__dict__.items():
            # access = 'master'
            found = False
            auth_level = ""
            for j in Sm.all_apis(None, True):
                if i == j["fname"]:
                    found = True
                    auth_level = (
                        "public"
                        if j in Sm._public_api
                        else "user"
                        if j in Sm._private_api
                        else "admin"
                        if j in Sm._admin_api
                        else "cli_only"
                    )
                    break
            if not found:
                continue
            name = i.replace("_", " ")
            api = i.replace("_", "\\_")
            ret += (
                f"\\subsubsection{{\\lstinline"
                f"[basicstyle=\\Large\\ttfamily]${name}$}}\n\n"
            )
            authstr = "(cli only)"
            if auth_level != "cli_only":
                authstr = f"| api: {api} | auth: {auth_level}"
            ret += self.func_to_sexy_box(f"cli: {name} {authstr}", v)
        return ret



"""
Sentinel class for Jaseci

Each sentinel has an id, name, timestamp and it's set of walkers.
"""
from jaseci.element.element import Element
from jaseci.utils.utils import logger, ColCodes as Cc, is_true
from jaseci.utils.id_list import IdList
from jaseci.jac.ir.jac_code import JacCode, jac_ir_to_ast
from jaseci.jac.interpreter.sentinel_interp import SentinelInterp
from jaseci.actor.walker import Walker
from jaseci.actor.architype import Architype


class Sentinel(Element, JacCode, SentinelInterp):
    """
    Sentinel class for Jaseci

    is_active is used to signify whether sentinel is ready to run walkers, i.e,
    register_code succeeded
    """

    def __init__(self, *args, **kwargs):
        self.version = None
        self.arch_ids = IdList(self)
        self.global_vars = {}
        self.testcases = []
        Element.__init__(self, *args, **kwargs)
        JacCode.__init__(self, code_ir=None)
        SentinelInterp.__init__(self)

    def reset(self):
        """Resets state of sentinel and unregister's code"""
        self.version = None
        self.global_vars = {}
        self.testcases = []
        self.arch_ids.destroy_all()
        JacCode.reset(self)
        SentinelInterp.reset(self)

    def refresh(self):
        super().refresh()
        self.ir_load()

    def register_code(self, text, dir="./", mode="default"):
        """
        Registers a program (set of walkers and architypes) written in Jac
        """
        self.reset()
        if mode == "ir":
            self.apply_ir(text)
        else:
            self.register(text, dir)
        if self.is_active:
            self.ir_load()
        return self.is_active

    def load_arch_defaults(self):
        self.arch_ids.add_obj(
            Architype(m_id=self._m_id, h=self._h, name="root", kind="node", parent=self)
        )
        self.arch_ids.add_obj(
            Architype(
                m_id=self._m_id, h=self._h, name="generic", kind="node", parent=self
            )
        )
        self.arch_ids.add_obj(
            Architype(
                m_id=self._m_id, h=self._h, name="generic", kind="edge", parent=self
            )
        )

    def ir_load(self):
        """
        Load walkers and architypes from IR
        """
        self.load_arch_defaults()
        self.run_start(self._jac_ast)

        if self.runtime_errors:
            logger.error(str(f"{self.name}: Runtime problem processing sentinel!"))
            self.is_active = False
        elif not self.arch_ids:
            logger.error(str(f"{self.name}: No walkers nor architypes created!"))
            self.is_active = False
        return self.is_active

    def register_architype(self, code):
        """Adds an architype based on jac code"""
        tree = self.parse_jac(code, start_rule="architype")
        if not tree:
            return None
        return self.load_architype(tree)

    def spawn_architype(self, name, kind=None, caller=None, is_async=None):
        """
        Spawns a new architype from registered architypes and adds to
        live walkers
        """
        src_arch = self.arch_ids.get_obj_by_name(name, kind=kind, silent=True)
        if not src_arch:
            logger.error(str(f"{self.name}: Unable to spawn {kind} architype {name}!"))
            return None
        src_arch.is_async = src_arch.is_async if is_async is None else is_true(is_async)
        if caller and caller._m_id != src_arch._m_id:
            new_arch = src_arch.duplicate()
            new_arch.set_master(caller._m_id)
            new_arch._jac_ast = src_arch._jac_ast
            if new_arch._jac_ast is None:
                new_arch.refresh()
            return new_arch
        else:
            return src_arch

    def run_architype(self, name, kind=None, caller=None, is_async=None):
        """
        Spawn, run, then destroy architype if m_id's are different
        """
        if caller is None:
            caller = self
        arch = self.spawn_architype(name, kind, caller, is_async)
        if arch is None:
            logger.error(
                str(f"{self.name}: Unable to spawn architype " f"{[name, kind]}!")
            )
            return None
        if arch.jid in self.arch_ids:
            return arch.run()
        else:
            ret = arch.run()
            Element.destroy(arch)
            return ret

    def get_arch_for(self, obj):
        """Returns the architype that matches object"""
        ret = self.arch_ids.get_obj_by_name(name=obj.name, kind=obj.kind)
        if ret is None:
            self.rt_error(f"Unable to find architype for {obj.name}, {obj.kind}")
        return ret

    def run_tests(self, detailed=False, silent=False):
        """
        Testcase schema
        testcase = {'title': kid[1].token_text(),
            'graph_ref': None, 'graph_block': None,
            'walker_ref': None, 'spawn_ctx': None,
            'assert_block': None, 'walker_block': None,
            'passed': None}
        """
        from pprint import pformat
        from time import time
        import sys
        import io

        num_failed = 0
        for i in self.testcases:
            screen_out = [sys.stdout, sys.stderr]
            buff_out = [io.StringIO(), io.StringIO()]
            destroy_set = []
            title = i["title"]
            if i["graph_ref"]:
                gph = self.run_architype(i["graph_ref"], kind="graph", caller=self)
            else:
                gph = Architype(
                    m_id=self._m_id,
                    h=self._h,
                    parent=self,
                    code_ir=jac_ir_to_ast(i["graph_block"]),
                )
                destroy_set.append(gph)
                gph = gph.run()
            if i["walker_ref"]:
                wlk = self.run_architype(
                    name=i["walker_ref"], kind="walker", caller=self
                )
            else:
                wlk = Walker(
                    m_id=self._m_id,
                    h=self._h,
                    parent=self,
                    code_ir=jac_ir_to_ast(i["walker_block"]),
                )
                destroy_set.append(wlk)
            wlk.prime(gph)
            if i["spawn_ctx"]:
                self.run_spawn_ctx(jac_ir_to_ast(i["spawn_ctx"]), wlk)

            stime = time()
            try:
                if not silent:
                    print(f"Testing {title}: ", end="")
                sys.stdout, sys.stderr = buff_out[0], buff_out[1]
                wlk.run()
                sys.stdout, sys.stderr = screen_out[0], screen_out[1]
                if i["assert_block"]:
                    wlk._loop_ctrl = None
                    wlk.scope_and_run(
                        jac_ir_to_ast(i["assert_block"]), run_func=wlk.run_code_block
                    )
                i["passed"] = True
                if not silent:
                    print(
                        f"[{Cc.TG}PASSED{Cc.EC} in {Cc.TY}{time()-stime:.2f}s{Cc.EC}]"
                    )
                if detailed and not silent:
                    print("REPORT: " + pformat(wlk.report))
            except Exception as e:
                sys.stdout, sys.stderr = screen_out[0], screen_out[1]
                i["passed"] = False
                num_failed += 1
                if not silent:
                    print(
                        f"[{Cc.TR}FAILED{Cc.EC} in {Cc.TY}{time()-stime:.2f}s{Cc.EC}]"
                    )
                    print(f"{e}")
            for i in destroy_set:  # FIXME: destroy set not complete
                i.destroy()
        num_tests = len(self.testcases)
        summary = {
            "tests": num_tests,
            "passed": num_tests - num_failed,
            "failed": num_failed,
            "success": num_tests and not num_failed,
        }
        if detailed:
            details = []
            for i in self.testcases:
                details.append(
                    {
                        "test": i["title"],
                        "passed": i["passed"],
                        "stdout": buff_out[0].getvalue(),
                        "stderr": buff_out[1].getvalue(),
                    }
                )
            summary["details"] = details
        return summary

    def destroy(self):
        """
        Destroys self from memory and persistent storage
        """
        for i in self.arch_ids.obj_list():
            i.destroy()
        super().destroy()



from os.path import dirname, join


def get_ver():
    with open(join(dirname(__file__), "VERSION")) as version_file:
        return version_file.read().strip()


__version__ = get_ver()
__creator__ = "Jason Mars"
__url__ = "https://jaseci.org"


def load_standard():
    import jaseci.actions.standard.net  # noqa
    import jaseci.actions.standard.rand  # noqa
    import jaseci.actions.standard.request  # noqa
    import jaseci.actions.standard.std  # noqa
    import jaseci.actions.standard.file  # noqa
    import jaseci.actions.standard.vector  # noqa
    import jaseci.actions.standard.date  # noqa
    import jaseci.actions.standard.jaseci  # noqa
    import jaseci.actions.standard.mail  # noqa
    import jaseci.actions.standard.task  # noqa


load_standard()



from enum import Enum


class ServiceState(Enum):
    FAILED = -2
    DISABLED = -1
    NOT_STARTED = 0
    STARTED = 1
    RUNNING = 2

    def is_ready(self):
        return self == ServiceState.NOT_STARTED

    def is_running(self):
        return self == ServiceState.RUNNING

    def has_failed(self):
        return self == ServiceState.FAILED



import io
import pstats
import cProfile
import pdb
import uuid
import importlib
import pkgutil
import logging
import types
import base64
import re
import json
import functools
import traceback
import inspect
from time import time
from datetime import datetime
from pprint import pformat


class ColCodes:
    TY = "\033[33m"
    TG = "\033[32m"
    TR = "\033[31m"
    EC = "\033[m"


def master_from_meta(meta):
    """Return master from meta in actions"""
    return meta["h"].get_obj(meta["m_id"], uuid.UUID(meta["m_id"]))


# Get an instance of a logger
def connect_logger_handler(target_logger, handler, level=logging.WARN):
    """Attaches standard formatting and adds handler to logger"""
    target_logger.setLevel(level)
    handler.setFormatter(
        logging.Formatter("%(asctime)s - %(levelname)s - %(funcName)s: %(message)s")
    )
    target_logger.addHandler(handler)


logger = logging.getLogger("core")
if len(logger.handlers) < 1:
    connect_logger_handler(logger, logging.StreamHandler(), logging.INFO)

app_logger = logging.getLogger("app")
if len(app_logger.handlers) < 1:
    connect_logger_handler(app_logger, logging.StreamHandler(), logging.INFO)


def log_var_out(val):
    """Print to log"""
    if not logging.getLogger("core").disabled:
        logger.info(pformat(val))


camel_to_snake_re = re.compile(r"(?<!^)(?=[A-Z])")


def camel_to_snake(name):
    return camel_to_snake_re.sub("_", name).lower()


def bp():
    pdb.set_trace()
    breakpoint()


def dummy_bp(inspect):
    traceback.print_stack()
    print(inspect)
    input()


def print_stack_to_log():
    tb = traceback.extract_stack()
    log_var_out(tb)


uuid_re = re.compile(
    "([a-f0-9]{8}-?[a-f0-9]{4}-?4[a-f0-9]{3}-?[89aAbB][a-f0-9]{3}-?[a-f0-9]{12})"
)


def is_urn(s: str):
    """Test if is uuid string in urn format"""
    return type(s) == str and s.startswith("urn:uuid:")


def is_jsonable(x):
    """Test if object can be json serialized"""
    try:
        json.dumps(x)
        return True
    except (TypeError, OverflowError):
        return False


def json_out(val):
    if type(val) == str:
        return val
    try:
        return json.dumps(val)
    except Exception:
        return val


def parse_str_token(s):
    return str(bytes(s, "utf-8").decode("unicode_escape")[1:-1])


def get_all_subclasses(cls):
    """Return list of all subclasses of cls"""
    return set(cls.__subclasses__()).union(
        [s for c in cls.__subclasses__() for s in get_all_subclasses(c)]
    )


def matching_fields(obj1, obj2):
    """
    Return list of matching member attributes in objects
    (non-private fields only)
    """
    matches = []
    for a in dir(obj1):
        if not a.startswith("_"):
            for b in dir(obj2):
                if a == b:
                    matches.append(a)
    return matches


def map_assignment_of_matching_fields(dest, source):
    """
    Assign the values of identical feild names from source to destination.
    """
    for i in matching_fields(dest, source):
        if type(getattr(source, i)) == uuid.UUID:
            setattr(dest, i, getattr(source, i).urn)
        elif type(getattr(source, i)) == datetime:
            setattr(dest, i, getattr(source, i).isoformat())
        elif not callable(getattr(dest, i)):
            setattr(dest, i, getattr(source, i))


obj_class_cache = {}


def build_class_dict(from_where):
    global obj_class_cache
    prefix = from_where.__name__ + "."
    for importer, modname, ispkg in pkgutil.iter_modules(from_where.__path__, prefix):
        if not ispkg:
            clsmembers = inspect.getmembers(
                importlib.import_module(modname), inspect.isclass
            )
            for cls, obj in clsmembers:
                obj_class_cache[camel_to_snake(cls)] = getattr(
                    importlib.import_module(modname), cls
                )
        else:
            if hasattr(from_where, modname.split(".")[-1]):
                build_class_dict(getattr(from_where, modname.split(".")[-1]))


def find_class_and_import(class_name, from_where):
    """
    Search for class through all core packages

    Classes assumed to have same name as module file
    """
    global obj_class_cache
    if not obj_class_cache.keys():
        build_class_dict(from_where)
    return obj_class_cache[class_name]


def copy_func(f, name=None):
    """
    Utility to duplicate function in python.
    Can be used to programatically add methods to classes
    """
    g = types.FunctionType(
        f.__code__,
        f.__globals__,
        name=f.__name__,
        argdefs=f.__defaults__,
        closure=f.__closure__,
    )
    g = functools.update_wrapper(g, f)
    g.__kwdefaults__ = f.__kwdefaults__
    g.__name__ = name if name else g.__name__
    return g


def b64decode_str(code):
    """Decode a base 64 encoded string"""
    try:
        code = base64.b64decode(code).decode()
    except UnicodeDecodeError:
        logger.error("Code encoding invalid!")
    return code


perf_prof = None


def perf_test_start():
    global perf_prof
    perf_prof = cProfile.Profile()
    perf_prof.enable()


def perf_test_stop():
    perf_prof.disable()
    s = io.StringIO()
    sortby = pstats.SortKey.CUMULATIVE
    ps = pstats.Stats(perf_prof, stream=s).sort_stats(sortby)
    ps.print_stats(100)
    print(s.getvalue())


class TestCaseHelper:
    """Helper to pretty print test results"""

    def setUp(self):
        self.logger_off()
        self.stime = time()
        return super().setUp()

    def tearDown(self):
        td = super().tearDown()
        result = (
            f"Time: {ColCodes.TY}{time()-self.stime:.3f} "
            + f'- {ColCodes.EC}{self.id().split(".")[-1]}: '
        )
        get_outcome = self.defaultTestResult()
        self._feedErrorsToResult(get_outcome, self._outcome.errors)
        if len(get_outcome.errors) or len(get_outcome.failures):
            result += f"{ColCodes.TR}[failed]{ColCodes.EC}"
        elif len(self._outcome.skipped):
            result += f"{ColCodes.TY}[skipped]{ColCodes.EC}"
        else:
            result += f"{ColCodes.TG}[passed]{ColCodes.EC}"

        print(result)
        self.logger_on()
        return td

    def logger_off(self):
        """Turn off logging output"""
        logging.getLogger("core").disabled = True
        logging.getLogger("app").disabled = True

    def logger_on(self):
        """Turn on logging output"""
        logging.getLogger("core").disabled = False
        logging.getLogger("app").disabled = False

    def is_logger_off(self):
        return logging.getLogger("core").disabled and logging.getLogger("app").disabled

    def log(self, *val):
        """Print to log"""
        is_off = self.is_logger_off()
        self.logger_on()
        for i in val:
            log_var_out(i)
        if is_off:
            self.logger_off()

    def stopper(self):
        """Force test to fail"""
        self.assertTrue(False)

    def perf_test_start(self):
        self.pr = cProfile.Profile()
        self.pr.enable()

    def perf_test_stop(self):
        self.pr.disable()
        s = io.StringIO()
        sortby = pstats.SortKey.CUMULATIVE
        ps = pstats.Stats(self.pr, stream=s).sort_stats(sortby)
        ps.print_stats(100)
        print(s.getvalue())


def is_true(val):
    return (
        val.lower() == "true"
        if type(val) is str
        else val is True  # is_async might be non bool
    )



import os
from unittest import TestCase

from jaseci.svc import MetaService
from jaseci.utils.utils import TestCaseHelper


class CoreTest(TestCaseHelper, TestCase):
    """Unit tests for Jac Core APIs"""

    fixture_src = __file__

    def setUp(self):
        super().setUp()
        self.meta = MetaService()
        self.smast = self.meta.build_super_master()
        self.mast = self.meta.build_master(h=self.smast._h)

    def tearDown(self):
        super().tearDown()

    def call(self, mast, pl):
        ret = mast.general_interface_to_api(api_name=pl[0], params=pl[1])
        return ret

    def load_jac(self, fn):
        with open(os.path.dirname(self.fixture_src) + "/fixtures/" + fn) as f:
            return f.read()



"""
Command line tool for Jaseci
"""

import functools
import json
import os
import pickle

import click
import requests
from click_shell import shell

from jaseci import __version__
from jaseci.element.super_master import SuperMaster
from jaseci.svc import MetaService
from jaseci.utils.utils import copy_func
from .book_tools import Book

session = None


def reset_state():
    global session
    session = {
        "filename": "js.session",
        "user": [MetaService(run_svcs=False).build_super_master(name="admin")],
        "mem-only": session["mem-only"] if session is not None else False,
        "connection": {"url": None, "token": None, "headers": None},
    }
    session["master"] = session["user"][0]


reset_state()


def is_connected():
    return session["connection"]["token"] and session["connection"]["url"]


def get_prompt():
    if is_connected():
        return "@jaseci > "
    else:
        return "jaseci > "


def get_intro():
    return f"Jaseci {__version__}\n" + "Starting Shell..."


@shell(prompt=get_prompt, intro=get_intro())
@click.option(
    "--filename", "-f", default="js.session", help="Specify filename for session state."
)
@click.option(
    "--mem-only", "-m", is_flag=True, help="Set true to not save file for session."
)
def jsctl(filename, mem_only):
    """
    The Jaseci Command Line Interface
    """
    global session
    if not mem_only and os.path.isfile(filename):
        with open(filename, "rb") as f:
            session = pickle.load(f)
    session["mem-only"] = mem_only
    session["filename"] = filename if not mem_only else None


def remote_api_call(payload, api_name):
    """
    Constructs and issues call to remote server
    NOTE: Untested
    """
    for i in SuperMaster.all_apis(None):
        if api_name == "_".join(i["groups"]):
            if i in SuperMaster._private_api:
                path = "/js/" + api_name
            elif i in SuperMaster._admin_api:
                path = "/js_admin/" + api_name
            elif i in SuperMaster._public_api:
                path = "/js_public/" + api_name
            break
    ret = requests.post(
        session["connection"]["url"] + path,
        json=payload,
        headers=session["connection"]["headers"],
    )
    if ret.status_code > 205:
        ret = f"Status Code Error {ret.status_code}"
    else:
        ret = ret.json()
    return ret


def resolve_none_type(kwargs):
    for i in kwargs.keys():
        if kwargs[i] == "None":
            kwargs[i] = None


def interface_api(api_name, is_public, is_cli_only, **kwargs):
    """
    Interfaces Master apis after processing arguments/parameters
    from cli
    """
    if "code" in kwargs and kwargs["code"]:
        if os.path.isfile(kwargs["code"]):
            with open(kwargs["code"], "r") as file:
                if (
                    api_name == "sentinel_register"
                    and "name" in kwargs
                    and kwargs["name"] == "default"
                ):
                    kwargs["name"] = kwargs["code"]
                kwargs["code"] = file.read()
        else:
            click.echo(f"Code file {kwargs['code']} not found!")
            return
    resolve_none_type(kwargs)
    if not is_cli_only and is_connected():
        out = remote_api_call(kwargs, api_name)
    elif is_public:
        out = session["master"].public_interface_to_api(kwargs, api_name)
    else:
        out = session["master"].general_interface_to_api(kwargs, api_name)
    if (
        isinstance(out, dict)
        and "report_custom" in out.keys()
        and out["report_custom"] is not None
    ):
        out = out["report_custom"]
    if isinstance(out, dict) or isinstance(out, list):
        out = json.dumps(out, indent=2)
    click.echo(out)
    if "output" in kwargs and kwargs["output"]:
        with open(kwargs["output"], "w") as f:
            f.write(out)
        click.echo(f'[saved to {kwargs["output"]}]')
    if not session["mem-only"]:
        with open(session["filename"], "wb") as f:
            pickle.dump(session, f)


def extract_api_tree():
    """
    Generates a tree of command group names and function
    signatures in leaves from API function names in Master
    """
    api_funcs = {}
    for i in session["master"].all_apis() + session["master"]._cli_api:
        # Build hierarchy of command groups
        api_root = api_funcs
        for j in i["groups"]:
            if j not in api_root.keys():
                api_root[j] = {}
            api_root = api_root[j]
        api_root["leaf"] = [
            "_".join(i["groups"]),
            i["sig"],
            i in session["master"]._public_api,
            i["doc"],
            i["cli_args"],
            i in session["master"]._cli_api,
        ]
    return api_funcs


def build_cmd(group_func, func_name, leaf):
    """
    Generates Click function with options for each command
    group and leaf signatures
    leaf is format: [api_name, func_sig, is_public, func_doc]
    """

    f = functools.partial(
        copy_func(interface_api, func_name),
        api_name=leaf[0],
        is_public=leaf[2],
        is_cli_only=leaf[5],
    )
    f.__name__ = func_name
    f.__doc__ = leaf[3]

    func_sig = leaf[1]
    cli_args = leaf[4]
    for i in func_sig.parameters.keys():
        if i == "self":
            continue
        p_default = func_sig.parameters[i].default
        p_type = func_sig.parameters[i].annotation
        if p_type not in [int, bool, float]:
            p_type = str
        if i in cli_args:
            f = click.argument(f"{i}", type=p_type)(f)
        elif p_default is not func_sig.parameters[i].empty:
            f = click.option(
                f"-{i}",
                default=p_default if p_default is None else p_type(p_default),
                required=False,
                type=p_type,
            )(f)
        else:
            f = click.option(f"-{i}", required=True, type=p_type)(f)
    # to file option to dump response to a file
    f = click.option(
        "--output",
        "-o",
        default="",
        required=False,
        type=str,
        help="Filename to dump output of this command call.",
    )(f)
    return group_func.command()(f)


def cmd_tree_builder(location, group_func=jsctl, cmd_str=""):
    """
    Generates Click command groups from API tree recursively
    """
    for i in location.keys():
        loc = location[i]
        if "leaf" in loc:
            build_cmd(group_func, i, loc["leaf"])
            continue
        else:
            f = copy_func(lambda: None, i)
            f.__doc__ = f'Group of `{(cmd_str + " " + i).lstrip()}` commands'
            new_func = group_func.group()(f)
        cmd_tree_builder(loc, new_func, cmd_str + " " + i)


@click.command(help="Command to log into live Jaseci server")
@click.argument("url", type=str, required=True)
@click.option(
    "--username",
    "-u",
    required=True,
    prompt=True,
    help="Username to be used for login.",
)
@click.password_option(help="Password to be used for login.", confirmation_prompt=False)
def login(url, username, password):
    url = url[:-1] if url[-1] == "/" else url
    payload = {"email": username, "password": password}
    try:
        r = requests.post(url + "/user/token/", data=payload).json()
    except Exception:
        r = {"error": "Invalid url, username, or password."}
    if "token" in r.keys():
        session["connection"]["token"] = r["token"]
        session["connection"]["url"] = url
        session["connection"]["headers"] = {"Authorization": "token " + r["token"]}
        click.echo(f"Token: {r['token']}\nLogin successful!")
    else:
        click.echo("Login failed!\n")
    if not session["mem-only"]:
        with open(session["filename"], "wb") as f:
            pickle.dump(session, f)


@click.command(help="Command to log out of live Jaseci server")
def logout():
    if session["connection"]["token"]:
        session["connection"]["token"] = None
        session["connection"]["url"] = None
        session["connection"]["headers"] = {}
        click.echo("Logout successful!")
    else:
        click.echo("You are not logged in!")


@click.command(help="Edit a file")
@click.argument("file", type=str, required=True)
def edit(file):
    click.edit(filename=file)


@click.command(help="List relevant files")
@click.option(
    "--all",
    "-a",
    is_flag=True,
    help="Flag for listing all files, not just relevant files",
)
def ls(all):
    for i in os.listdir():
        if all:
            click.echo(f"{i}")
        else:
            click.echo(f"{i}") if i.endswith(".jac") or i.endswith(".dot") else False


@click.command(help="Clear terminal")
def clear():
    click.clear()


@click.command(help="Reset jsctl (clears state)")
def reset():
    reset_state()
    click.echo("Jaseci State Cleared!")


@click.command(help="Internal book generation tools")
@click.argument("op", type=str, default="cheatsheet", required=True)
@click.option(
    "--output",
    "-o",
    default="",
    required=False,
    type=str,
    help="Filename to dump output of this command call.",
)
def booktool(op, output):
    out = ""
    if op == "cheatsheet":
        out = f"{Book().bookgen_api_cheatsheet(extract_api_tree())}"
    elif op == "stdlib":
        out = Book().bookgen_std_library()
    elif op == "classes":
        out = Book().bookgen_api_spec()
    click.echo(out)
    if output:
        with open(output, "w") as f:
            f.write(out)
        click.echo(f"[saved to {output}]")


jsctl.add_command(login)
jsctl.add_command(logout)
jsctl.add_command(edit)
jsctl.add_command(ls)
jsctl.add_command(clear)
jsctl.add_command(reset)
jsctl.add_command(booktool)
cmd_tree_builder(extract_api_tree())


def main():
    jsctl()


if __name__ == "__main__":
    main()



from jaseci.utils.test_core import CoreTest


class FileLibTest(CoreTest):

    fixture_src = __file__

    def test_json_dump(self):
        ret = self.call(
            self.mast,
            ["sentinel_register", {"code": self.load_jac("file_stuff.jac")}],
        )
        ret = self.call(self.mast, ["walker_run", {"name": "pack_it"}])
        self.assertEqual(ret["report"][0], {"hello": 5})



from jaseci.jac.ir.passes.prune_pass import PrunePass
from jaseci.jac.ir.passes.printer_pass import PrinterPass
from jaseci.jac.ir.passes.stats_pass import StatsPass


def multi_pass_optimizer(jac_ast, debug_out=False, print_walk=False):
    if debug_out:
        if print_walk:
            PrinterPass(jac_ast).run()
        StatsPass(jac_ast).run()
    PrunePass(jac_ast).run()
    if debug_out:
        StatsPass(jac_ast).run()
